import numpy as np
import logging
import uuid
from copy import deepcopy
from openpnm.core import (
    LabelMixin,
    ParserMixin,
    ModelsMixin2
)
from openpnm.utils import (
    Workspace,
    SettingsAttr,
    PrintableList,
    PrintableDict,
    Docorator,
    get_printable_props,
    get_printable_labels,
)


docstr = Docorator()
logger = logging.getLogger(__name__)
ws = Workspace()


__all__ = [
    'Base2',
    'Domain',
]


@docstr.get_sections(base='BaseSettings', sections=docstr.all_sections)
@docstr.dedent
class BaseSettings:
    r"""
    The default settings to use on instance of Base

    Parameters
    ----------
    uuid : str
        A universally unique identifier for the object to keep things straight

    """
    default_domain = 'domain_1'


@docstr.get_sections(base='Base', sections=['Parameters'])
@docstr.dedent
class Base2(dict):
    r"""
    A subclassed dictionary used for storing data

    Parameters
    ----------
    network : dict
        An OpenPNM Network object, which is a subclass of a dict
    """

    def __new__(cls, *args, **kwargs):
        instance = super(Base2, cls).__new__(cls, *args, **kwargs)
        # It is necessary to set the SettingsAttr here since some classes
        # use it before calling super.__init__()
        instance.settings = SettingsAttr()
        instance.settings['uuid'] = str(uuid.uuid4())
        return instance

    def __init__(self, network=None, project=None, name='obj_?'):
        super().__init__()
        # Add default settings
        self.settings._update(BaseSettings())
        # Add parameters attr
        self._params = PrintableDict(key="Parameters", value="Value")
        # Associate with project
        if (network is None) and (project is None):
            project = ws.new_project()
        elif project is None:
            project = network.project
        project.append(self)
        self.name = name

    def __eq__(self, other):
        return hex(id(self)) == hex(id(other))

    def __repr__(self):  # pragma: no cover
        module = self.__module__
        module = ".".join([x for x in module.split(".") if not x.startswith("_")])
        cname = self.__class__.__name__
        return f'{self.name} : <{module}.{cname} at {hex(id(self))}>'

    def __setitem__(self, key, value):
        if value is None:
            return

        # Intercept parameters
        if key.startswith('param'):
            _, key = key.split('.', 1)
            self._params[key] = value
            return

        if not (key.startswith('pore.') or key.startswith('throat.')):
            raise Exception("All dict names must start with pore, throat, or param")

        # Intercept @ symbol
        if '@' in key:
            element, prop = key.split('@')[0].split('.', 1)
            domain = key.split('@')[1]
            locs = super().__getitem__(f'{element}.{domain}')
            try:
                vals = self[f'{element}.{prop}']
                vals[locs] = value
                self[f'{element}.{prop}'] = vals
            except KeyError:
                value = np.array(value)
                temp = self._initialize_empty_array_like(value, element)
                self.__setitem__(f'{element}.{prop}', temp)
                self[f'{element}.{prop}'][locs] = value
            return

        element, prop = key.split('.', 1)
        # Catch dictionaries and break them up
        if isinstance(value, dict):
            for k, v in value.items():
                self[f'{key}.{k}'] = v
            return

        # Enforce correct dict naming
        if element not in ['pore', 'throat']:
            raise Exception('All keys must start with either pore or throat')
        # Convert value to ndarray
        if not isinstance(value, np.ndarray):
            value = np.array(value, ndmin=1)
        # Skip checks for coords and conns
        if key in ['pore.coords', 'throat.conns']:
            self.update({key: value})
            return
        # Finally write data
        if self._count(element) is None:
            self.update({key: value})  # If length not defined, do it
        elif value.shape[0] == 1:  # If value is scalar
            value = np.ones((self._count(element), ), dtype=value.dtype)*value
            self.update({key: value})
        elif np.shape(value)[0] == self._count(element):
            self.update({key: value})
        else:
            raise Exception('Provided array is wrong length for ' + key)

    def __getitem__(self, key):
        # If key is a just a numerical value, then kick it directly back.
        # This allows one to do either value='pore.blah' or value=1.0 in
        # pore-scale models
        if not isinstance(key, str):
            return key

        if key.startswith('param'):
            _, key = key.split('.', 1)
            try:
                return self._params[key]
            except KeyError:
                return self.network._params[key]

        # If key contains an @ symbol then return a subset of values at the
        # requested locations, by recursively calling __getitem__
        if '@' in key:
            element, prop = key.split('@')[0].split('.', 1)
            domain = key.split('@')[1]
            if f'{element}.{domain}' not in self.keys():
                raise KeyError(key)
            locs = self[f'{element}.{domain}']
            vals = self[f'{element}.{prop}']
            return vals[locs]

        try:
            return super().__getitem__(key)
        except KeyError:
            # If key is object's name or all, return ones
            if key.split('.', 1)[-1] in [self.name, 'all']:
                element, prop = key.split('.', 1)
                vals = np.ones(self._count(element), dtype=bool)
                return vals
            else:
                vals = {}  # Gather any arrays into a dict
                for k in self.keys():
                    if k.startswith(f'{key}.'):
                        vals.update({k.replace(f'{key}.', ''): self[k]})
                if len(vals) > 0:
                    return vals
                else:
                    raise KeyError(key)

    def __delitem__(self, key):
        try:
            super().__delitem__(key)
        except KeyError:
            d = self[key]  # If key is a nested dict, get all values
            for item in d.keys():
                super().__delitem__(f'{key}.{item}')

    def pop(self, *args):
        r"""
        """
        v = super().pop(*args)
        if v is None:
            try:
                d = self[args[0]]
                v = {}
                for item in d.keys():
                    key = f'{args[0]}.{item}'
                    v[key] = super().pop(key)
            except KeyError:
                pass
        return v

    def clear(self, mode=None):
        r"""
        Clears or deletes certain things from object. If no arguments are provided
        it defaults to the normal `dict` behavior.

        Parameters
        ----------
        mode : str
            Controls which things are to be deleted. Options are:

            =========== ============================================================
            `mode`        Description
            =========== ============================================================
            'props'     Deletes all pore and throat properties (i.e numerical data)
                        in the object's dictionary (except 'pore.coords' and
                        'throat.conns' if it is a network object).
            'labels'    Deletes all labels (i.e. boolean data) in the object's
                        dictionary.
            'models'    Delete are pore and throat properties that were produced
                        by a pore-scale model.
            =========== ============================================================

        """
        if mode is None:
            super().clear()
        else:
            if isinstance(mode, str):
                mode = [mode]
            if 'props' in mode:
                for item in self.props():
                    if item not in ['pore.coords', 'throat.conns']:
                        del self[item]
            if 'labels' in mode:
                for item in self.labels():
                    if item not in ['pore.'+self.name, 'throat.'+self.name]:
                        del self[item]
            if 'models' in mode:
                for item in self.models.keys():
                    _ = self.pop(item.split('@')[0], None)

    def keys(self, mode=None):
        r"""
        An overloaded version of ``keys`` that optionally accepts a ``mode``

        Parameters
        ----------
        mode : str
            If given, optionally, it controls which type of keys are returned.
            Options are:

            ========== =======================================================
            mode       description
            ========== =======================================================
            props      Returns only keys that contain numerical arrays
            labels     Returns only keys that contain boolean arrays
            models     Returns only keys that were generated by a pore-scale
                       model
            constants  Returns only keys are were *not* generated by a pore-
                       scale model
            ========== =======================================================

        """
        if mode is None:
            return super().keys()
        else:
            if isinstance(mode, str):
                mode = [mode]
            vals = set()
            if 'props' in mode:
                for item in self.props():
                    vals.add(item)
            if 'labels' in mode:
                for item in self.labels():
                    vals.add(item)
            if 'models' in mode:
                for item in self.models.keys():
                    propname = item.split('@')[0]
                    if propname in self.keys():
                        vals.add(propname)
            if 'constants' in mode:
                vals = vals.union(set(self.props()))
                for item in self.models.keys():
                    propname = item.split('@')[0]
                    if propname in vals:
                        vals.remove(propname)
            return PrintableList(vals)

    def _set_name(self, name, validate=True):
        if not hasattr(self, '_name'):
            self._name = ''
        old_name = self._name
        if name == old_name:
            return
        name = self.project._generate_name(name)
        self._name = name

    def _get_name(self):
        try:
            return self._name
        except AttributeError:
            return None

    name = property(_get_name, _set_name)

    def _get_project(self):
        for proj in list(ws.values()):
            if self in proj:
                return proj

    project = property(fget=_get_project)

    def _set_settings(self, settings):
        self._settings = deepcopy(settings)

    def _get_settings(self):
        if self._settings is None:
            self._settings = SettingsAttr()
        return self._settings

    def _del_settings(self):
        self._settings = None

    settings = property(fget=_get_settings, fset=_set_settings, fdel=_del_settings)

    @property
    def network(self):
        r"""
        Shortcut to retrieve a handle to the network object associated with the
        calling object
        """
        return self.project.network

    @property
    def params(self):
        r"""
        This attribute stores 'scalar' data that can be used by pore-scale models.
        For instance, if a model calls for `temperature` you can specify
        `pore.temperature` if every pore might have a different value, or
        `param.temperature` if a single value prevails everywhere.
        """
        return self._params

    def _count(self, element):
        if element == 'pore':
            try:
                return self['pore.coords'].shape[0]
            except KeyError:
                for k, v in self.items():
                    if k.startswith('pore.'):
                        return v.shape[0]
        elif element == 'throat':
            try:
                return self['throat.conns'].shape[0]
            except KeyError:
                for k, v in self.items():
                    if k.startswith('throat.'):
                        return v.shape[0]

    @property
    def Nt(self):
        r"""
        Shortcut to retrieve the number of throats in the domain
        """
        return self._count('throat')

    @property
    def Np(self):
        r"""
        Shortcut to retrieve the number of pores in the domain
        """
        return self._count('pore')

    @property
    def Ts(self):
        r"""
        Shortcut to retrieve the indices of *all* throats
        """
        return np.arange(self._count('throat'))

    @property
    def Ps(self):
        r"""
        Shortcut to retrieve the indices of *all* pores
        """
        return np.arange(self._count('pore'))

    def _tomask(self, element, indices):
        return self.to_mask(**{element+'s': indices})

    def to_mask(self, pores=None, throats=None):
        r"""
        Generates a boolean mask with `True` values in the given locations

        Parameters
        ----------
        pores : array_like
            The pore indices where `True` values will be placed. If `pores` is
            given the `throats` is ignored.
        throats : array_like
            The throat indices where `True` values will be placed. If `pores` is
            given the `throats` is ignored.

        Returns
        -------
        mask : ndarray, boolean
            A boolean array of length Np is `pores` was given or Nt if
            `throats` was given.

        """
        if pores is not None:
            indices = np.array(pores, ndmin=1)
            N = self.Np
        elif throats is not None:
            indices = np.array(throats, ndmin=1)
            N = self.Nt
        else:
            raise Exception('Must specify either pores or throats')
        mask = np.zeros((N, ), dtype=bool)
        mask[indices] = True
        return mask

    def to_indices(self, mask):
        r"""
        Converts a boolean mask to pore or throat indices

        Parameters
        ----------
        mask : ndarray
            A boolean mask with `True` values indicating either pore or
            throat indices. This array must either be Nt or Np long, otherwise
            an Exception is raised.

        Returns
        -------
        indices : ndarray
            An array containing numerical indices of where `mask` was `True`.

        Notes
        -----
        This function is equivalent to just calling `np.where(mask)[0]` but
        does check to ensure that `mask` is a valid length.
        """
        mask = np.array(mask, dtype=bool)
        if mask.shape[0] not in [self.Np, self.Nt]:
            raise Exception('Mask must be either Nt or Np long')
        return np.where(mask)[0]

    def props(self, element=['pore', 'throat']):
        r"""
        Retrieves a list of keys that contain numerical data (i.e. "properties")

        Parameters
        ----------
        element : str, list of strings
            Indicates whether `'pore'` or `'throat'` properties should be returned.
            The default is `['pore', 'throat']`, so both are returned.

        Returns
        -------
        props : list of strings
            The names of all dictionary keys on the object that contain
            numerical data.
        """
        if element is None:
            element = ['pore', 'throat']
        if isinstance(element, str):
            element = [element]
        props = []
        for k, v in self.items():
            el, prop = k.split('.', 1)
            if (el in element) and (v.dtype != bool) and not prop.startswith('_'):
                props.append(k)
        props = sorted(props)
        props = PrintableList(props)
        return props

    def interpolate_data(self, propname, mode='mean'):
        r"""
        Generates an array of the requested pore/throat data by interpolating
        the neighboring throat/pore data.

        Parameters
        ----------
        propname : str
            The data to be generated.
        mode : str
            Dictate how the interpolation is done. Options are 'mean', 'min',
            and 'max'.

        Returns
        -------
        data : ndarray
            An ndarray containing the interpolated data.  E.g. Requesting
            'throat.temperature' will read the values of 'pore.temperature'
            in each of the neighboring pores and compute the average
            (if `mode='mean'`).
        """
        from openpnm.models.misc import from_neighbor_throats, from_neighbor_pores
        element, prop = propname.split('.', 1)
        if element == 'throat':
            if self['pore.'+prop].dtype == bool:
                raise Exception('The requested datatype is boolean, cannot interpolate')
            values = from_neighbor_pores(self, prop='pore.'+prop, mode=mode)
        elif element == 'pore':
            if self['throat.'+prop].dtype == bool:
                raise Exception('The requested datatype is boolean, cannot interpolate')
            values = from_neighbor_throats(self, prop='throat.'+prop, mode=mode)
        return values

    def get_conduit_data(self, propname):
        r"""
        Fetches an Nt-by-3 array of the requested property

        Parameters
        ----------
        propname : str
            The dictionary key of the property to fetch.

        Returns
        -------
        data : ndarray
            An Nt-by-3 array with each column containing the requrested data
            for pore1, throat, and pore2 respectively.

        """
        poreprop = 'pore.' + propname.split('.', 1)[-1]
        throatprop = 'throat.' + propname.split('.', 1)[-1]
        conns = self.network.conns
        try:
            T = self[throatprop]
            if T.ndim > 1:
                raise Exception(f'{throatprop} must be a single column wide')
        except KeyError:
            T = np.ones([self.Nt, ], dtype=float)*np.nan
        try:
            P1, P2 = self[poreprop][conns.T]
        except KeyError:
            P1 = np.ones([self.Nt, ], dtype=float)*np.nan
            P2 = np.ones([self.Nt, ], dtype=float)*np.nan
        vals = np.vstack((P1, T, P2)).T
        if np.isnan(vals).sum() == vals.size:
            raise KeyError(f'{propname} not found')
        return vals

    def __str__(self):  # pragma: no cover
        hr = '―' * 78
        lines = ''
        lines += '\n' + "═"*78 + '\n' + self.__repr__() + '\n' + hr + '\n'
        lines += get_printable_props(self)
        lines += '\n' + hr + '\n'
        lines += get_printable_labels(self)
        lines += '\n' + hr
        return lines

    def _initialize_empty_array_like(self, value, element):
        element = element.split('.', 1)[0]
        value = np.array(value)
        if value.dtype == bool:
            temp = np.zeros([self._count(element), *value.shape[1:]],
                            dtype=bool)
        else:
            temp = np.zeros([self._count(element), *value.shape[1:]],
                            dtype=float)*np.nan
        return temp


class Domain(ParserMixin, LabelMixin, ModelsMixin2, Base2):
    r"""
    The main class used for Network, Phase and Algorithm objects.

    This class inherits from ``Base2``, but also has several mixins for
    added functionality.

    """
    ...



from openpnm._skgraph import tools
from openpnm._skgraph import queries
from openpnm._skgraph import operations


__all__ = [
    'find_neighbor_sites',
    'find_neighbor_bonds',
    'find_connected_sites',
    'find_connecting_bonds',
    'istriu',
    'istril',
    'istriangular',
    'issymmetric',
    'tri_to_am',
    'vor_to_am',
    'conns_to_am',
    'drop_sites',
]


def find_neighbor_sites(sites, **kwargs):
    return queries.find_neighbor_nodes(inds=sites, **kwargs)


find_neighbor_sites.__doc__ = queries.find_neighbor_nodes.__doc__


def find_neighbor_bonds(sites, **kwargs):
    return queries.find_neighbor_edges(inds=sites, **kwargs)


find_neighbor_bonds.__doc__ = queries.find_neighbor_edges.__doc__


def find_connected_sites(bonds, **kwargs):
    return queries.find_connected_nodes(inds=bonds, **kwargs)


find_connected_sites.__doc__ = queries.find_connected_nodes.__doc__


def find_connecting_bonds(sites, **kwargs):
    return queries.find_connecting_edges(inds=sites, **kwargs)


find_connecting_bonds.__doc__ = queries.find_connecting_edges.__doc__


def istriu(am):
    return tools.istriu(am)


istriu.__doc__ = tools.istriu.__doc__


def istril(am):
    return tools.istril(am)


istril.__doc__ = tools.istril.__doc__


def istriangular(am):
    return tools.istriangular(am)


istriangular.__doc__ = tools.istriangular.__doc__


def issymmetric(am):
    return tools.issymmetric(am)


issymmetric.__doc__ = tools.issymmetric.__doc__


def tri_to_am(tri):
    return tools.tri_to_am(tri=tri)


tri_to_am.__doc__ = tools.tri_to_am.__doc__


def vor_to_am(vor):
    return tools.vor_to_am(vor=vor)


vor_to_am.__doc__ = tools.vor_to_am.__doc__


def conns_to_am(*args, **kwargs):
    return tools.conns_to_am(*args, **kwargs)


conns_to_am.__doc__ = tools.conns_to_am.__doc__


def drop_sites(am, sites):
    return operations.drop_nodes_from_am(inds=sites, am=am)


drop_sites.__doc__ = operations.drop_nodes_from_am.__doc__



import numpy as _np
from openpnm.models.geometry import _geodocs


__all__ = ["cylinder",
           "cuboid",
           "rectangle",
           "extrusion",
           "lens",
           "pendular_ring"]


@_geodocs
def cylinder(
    network,
    throat_diameter='throat.diameter',
    throat_length='throat.length',
):
    r"""
    Calculate throat volume assuing a cylindrical shape

    Parameters
    ----------
    %(network)s
    %(Dt)s
    %(Lt)s

    Returns
    -------
    volumes : ndarray
        A numpy ndarray containing throat volume values

    Notes
    -----
    This models does not account for the volume reprsented by the
    intersection of the throat with a spherical pore body.  Use the ``lens``
    or ``pendular_ring`` models in addition to this one to account for this
    volume.

    """
    leng = network[throat_length]
    diam = network[throat_diameter]
    value = _np.pi/4*leng*diam**2
    return value


@_geodocs
def cuboid(
    network,
    throat_diameter='throat.diameter',
    throat_length='throat.length',
):
    r"""
    Calculate throat volume assuing a square cross-section

    Parameters
    ----------
    %(network)s
    %(Dt)s
    %(Lt)s

    Returns
    -------

    Notes
    -----
    At present this models does NOT account for the volume reprsented by the
    intersection of the throat with a spherical pore body.

    """
    leng = network[throat_length]
    diam = network[throat_diameter]
    value = leng*diam**2
    return value


@_geodocs
def rectangle(
    network,
    throat_diameter='throat.diameter',
    throat_length='throat.length',
):
    r"""
    Calculate throat volume assuing a rectangular shape

    Parameters
    ----------
    %(network)s
    %(Dt)s
    %(Lt)s

    Returns
    -------

    Notes
    -----
    At present this models does NOT account for the volume reprsented by the
    intersection of the throat with a spherical pore body.
    """
    return network[throat_length] * network[throat_diameter]


@_geodocs
def extrusion(
    network,
    throat_length='throat.length',
    throat_area='throat.cross_sectional_area',
):
    r"""
    Calculate throat volume from the throat area and the throat length. This
    method is useful for abnormal shaped throats.

    Parameters
    ----------
    %(network)s
    %(Lt)s
    %(At)s

    Returns
    -------

    Notes
    -----
    At present this models does NOT account for the volume reprsented by the
    intersection of the throat with a spherical pore body.

    """
    leng = network[throat_length]
    area = network[throat_area]
    value = leng*area
    return value


@_geodocs
def lens(
    network,
    throat_diameter='throat.diameter',
    pore_diameter='pore.diameter',
):
    r"""
    Calculates the volume residing the hemispherical caps formed by the
    intersection between cylindrical throats and spherical pores.

    This volume should be subtracted from throat volumes if the throat lengths
    were found using throat end points.

    Parameters
    ----------
    %(network)s
    %(Dt)s
    %(Dp)s

    Returns
    -------

    Notes
    -----
    This model does not consider the possibility that multiple throats might
    overlap in the same location which could happen if throats are large and
    connectivity is random.

    See Also
    --------
    pendular_ring
    """
    conns = network['throat.conns']
    Rp = network[pore_diameter]/2
    Rt = network[throat_diameter]/2
    a = _np.atleast_2d(Rt).T
    q = _np.arcsin(a/Rp[conns])
    b = Rp[conns]*_np.cos(q)
    h = Rp[conns] - b
    V = 1/6*_np.pi*h*(3*a**2 + h**2)
    return _np.sum(V, axis=1)


@_geodocs
def pendular_ring(
    network,
    throat_diameter='throat.diameter',
    pore_diameter='pore.diameter',
):
    r"""
    Calculates the volume of the pendular rings residing between the end of
    a cylindrical throat and spherical pores that are in contact but not
    overlapping.

    This volume should be added to the throat volume if the throat length was
    found as the center-to-center distance less the pore radii.

    Parameters
    ----------
    %(network)s
    %(Dt)s
    %(Dp)s

    Returns
    -------

    Notes
    -----
    This model does not consider the possibility that multiple throats might
    overlap in the same location which could happen if throats are large and
    connectivity is random.

    See Also
    --------
    lens
    """
    conns = network['throat.conns']
    Rp = network[pore_diameter]/2
    Rt = network[throat_diameter]/2
    a = _np.atleast_2d(Rt).T
    q = _np.arcsin(a/Rp[conns])
    b = Rp[conns]*_np.cos(q)
    h = Rp[conns] - b
    Vlens = 1/6*_np.pi*h*(3*a**2 + h**2)
    Vcyl = _np.pi*(a)**2*h
    V = Vcyl - Vlens
    return _np.sum(V, axis=1)



import os
import numpy as np
import openpnm as op
import pandas as pd


def project_to_paraview(project, filename):  # pragma: no cover
    r"""
    Exports an OpenPNM network to a paraview state file.

    Parameters
    ----------
    network : Network
        The network containing the desired data.
    filename : str
        Path to saved .vtp file.

    Notes
    -----
    Outputs a pvsm file that can be opened in Paraview. The pvsm file will
    be saved with the same name as .vtp file

    """
    try:
        import paraview.simple
    except ModuleNotFoundError:
        msg = (
            "The paraview python bindings must be installed using "
            "conda install -c conda-forge paraview, however this may require"
            " using a virtualenv since conflicts with other packages are common."
            " This is why it is not explicitly included as a dependency in"
            " OpenPNM."
        )
        raise ModuleNotFoundError(msg)
    paraview.simple._DisableFirstRenderCameraReset()
    file = os.path.splitext(filename)[0]
    network = project.network
    x, y, z = np.ptp(network.coords, axis=0)
    if sum(op.topotools.dimensionality(network)) == 2:
        zshape = 0
        xshape = y
        yshape = x
    elif sum(op.topotools.dimensionality(network)) == 3:
        zshape = x
        xshape = z
        yshape = y
    maxshape = max(xshape, yshape, zshape)
    shape = np.array([xshape, yshape, zshape])
    # Create a new 'XML PolyData Reader'
    Path = os.getcwd() + "\\" + file + '.vtp'
    net_vtp = paraview.simple.XMLPolyDataReader(FileName=[Path])
    p = op.io.Dict.to_dict(project=project, element=['pore'],
                           flatten=False,
                           categorize_by=['data' 'object'])
    # The folloing lines are bit complex but avoid using the flatdict lib
    p = pd.json_normalize(p, sep='.').to_dict(orient='records')[0]
    for k in list(p.keys()):
        p[k.replace('.', ' | ')] = p.pop(k)
    t = op.io.Dict.to_dict(project=project, element=['throat'],
                           flatten=False,
                           categorize_by=['data' 'object'])
    t = pd.json_normalize(t, sep='.').to_dict(orient='records')[0]
    for k in list(t.keys()):
        t[k.replace('.', '/')] = t.pop(k)
    # Now write the p and t dictionaries
    net_vtp.CellArrayStatus = t.keys()
    net_vtp.PointArrayStatus = p.keys()
    # Get active view
    render_view = paraview.simple.GetActiveViewOrCreate('RenderView')
    # Uncomment following to set a specific view size
    # render_view.ViewSize = [1524, 527]
    # Get layout
    _ = paraview.simple.GetLayout()
    # Show data in view
    net_vtp_display = paraview.simple.Show(
        net_vtp, render_view, 'GeometryRepresentation'
    )
    # Trace defaults for the display properties.
    net_vtp_display.Representation = 'Surface'
    net_vtp_display.ColorArrayName = [None, '']
    net_vtp_display.OSPRayScaleArray = [
        f'network | {network.name} | labels | pore.all']
    net_vtp_display.OSPRayScaleFunction = 'PiecewiseFunction'
    net_vtp_display.SelectOrientationVectors = 'None'
    net_vtp_display.ScaleFactor = (maxshape-1)/10
    net_vtp_display.SelectScaleArray = 'None'
    net_vtp_display.GlyphType = 'Arrow'
    net_vtp_display.GlyphTableIndexArray = 'None'
    net_vtp_display.GaussianRadius = (maxshape-1)/200
    net_vtp_display.SetScaleArray = \
        ['POINTS', 'network | ' + network.name + ' | labels | pore.all']
    net_vtp_display.ScaleTransferFunction = 'PiecewiseFunction'
    net_vtp_display.OpacityArray = \
        ['POINTS', 'network | ' + network.name + ' | labels | pore.all']
    net_vtp_display.OpacityTransferFunction = 'PiecewiseFunction'
    net_vtp_display.DataAxesGrid = 'GridAxesRepresentation'
    net_vtp_display.PolarAxes = 'PolarAxesRepresentation'
    # Init the 'PiecewiseFunction' selected for 'ScaleTransferFunction'
    net_vtp_display.ScaleTransferFunction.Points = [1, 0, 0.5, 0, 1, 1, 0.5, 0]
    # Init the 'PiecewiseFunction' selected for 'OpacityTransferFunction'
    net_vtp_display.OpacityTransferFunction.Points = [1, 0, 0.5, 0, 1, 1, 0.5, 0]
    # Reset view to fit data
    render_view.ResetCamera()
    # Get the material library
    _ = paraview.simple.GetMaterialLibrary()
    # Update the view to ensure updated data information
    render_view.Update()
    # Create a new 'Glyph'
    glyph = paraview.simple.Glyph(Input=net_vtp, GlyphType='Arrow')
    glyph.OrientationArray = ['POINTS', 'No orientation array']
    glyph.ScaleArray = ['POINTS', 'No scale array']
    glyph.ScaleFactor = 1
    glyph.GlyphTransform = 'Transform2'
    # Properties modified on glyph
    glyph.GlyphType = 'Sphere'
    glyph.ScaleArray = \
        ['POINTS', 'network | ' + network.name + ' | properties | pore.diameter']
    # Show data in view
    glyph_display = paraview.simple.Show(
        glyph, render_view, 'GeometryRepresentation')
    # Trace defaults for the display properties.
    glyph_display.Representation = 'Surface'
    glyph_display.ColorArrayName = [None, '']
    glyph_display.OSPRayScaleArray = 'Normals'
    glyph_display.OSPRayScaleFunction = 'PiecewiseFunction'
    glyph_display.SelectOrientationVectors = 'None'
    glyph_display.ScaleFactor = (maxshape - 1) / 10
    glyph_display.SelectScaleArray = 'None'
    glyph_display.GlyphType = 'Arrow'
    glyph_display.GlyphTableIndexArray = 'None'
    glyph_display.GaussianRadius = (maxshape - 1) / 200
    glyph_display.SetScaleArray = ['POINTS', 'Normals']
    glyph_display.ScaleTransferFunction = 'PiecewiseFunction'
    glyph_display.OpacityArray = ['POINTS', 'Normals']
    glyph_display.OpacityTransferFunction = 'PiecewiseFunction'
    glyph_display.DataAxesGrid = 'GridAxesRepresentation'
    glyph_display.PolarAxes = 'PolarAxesRepresentation'
    # Init the 'PiecewiseFunction' selected for 'ScaleTransferFunction'
    glyph_display.ScaleTransferFunction.Points = [-0.97, 0, 0.5, 0, 0.97, 1, 0.5, 0]
    # Init the 'PiecewiseFunction' selected for 'OpacityTransferFunction'
    glyph_display.OpacityTransferFunction.Points = [-0.97, 0, 0.5, 0, 0.97, 1, 0.5, 0]
    # Update the view to ensure updated data information
    render_view.Update()
    # Set active source
    paraview.simple.SetActiveSource(net_vtp)
    # Create a new 'Shrink'
    shrink1 = paraview.simple.Shrink(Input=net_vtp)
    # Properties modified on shrink1
    shrink1.ShrinkFactor = 1.0
    # Show data in view
    shrink_display = paraview.simple.Show(
        shrink1, render_view, 'UnstructuredGridRepresentation')
    # Trace defaults for the display properties.
    shrink_display.Representation = 'Surface'
    shrink_display.ColorArrayName = [None, '']
    shrink_display.OSPRayScaleArray = \
        ['network | ' + network.name + ' | labels | pore.all']
    shrink_display.OSPRayScaleFunction = 'PiecewiseFunction'
    shrink_display.SelectOrientationVectors = 'None'
    shrink_display.ScaleFactor = (maxshape-1)/10
    shrink_display.SelectScaleArray = 'None'
    shrink_display.GlyphType = 'Arrow'
    shrink_display.GlyphTableIndexArray = 'None'
    shrink_display.GaussianRadius = (maxshape-1)/200
    shrink_display.SetScaleArray = \
        ['POINTS', 'network | ' + network.name + ' | labels | pore.all']
    shrink_display.ScaleTransferFunction = 'PiecewiseFunction'
    shrink_display.OpacityArray = \
        ['POINTS', 'network | ' + network.name + ' | labels | pore.all']
    shrink_display.OpacityTransferFunction = 'PiecewiseFunction'
    shrink_display.DataAxesGrid = 'GridAxesRepresentation'
    shrink_display.PolarAxes = 'PolarAxesRepresentation'
    shrink_display.ScalarOpacityUnitDistance = 1.0349360947089783
    # Init the 'PiecewiseFunction' selected for 'ScaleTransferFunction'
    shrink_display.ScaleTransferFunction.Points = [1, 0, 0.5, 0, 1, 1, 0.5, 0]
    # Init the 'PiecewiseFunction' selected for 'OpacityTransferFunction'
    shrink_display.OpacityTransferFunction.Points = [1, 0, 0.5, 0, 1, 1, 0.5, 0]
    # Hide data in view
    paraview.simple.Hide(net_vtp, render_view)
    # Update the view to ensure updated data information
    render_view.Update()
    # Create a new 'Cell Data to Point Data'
    cellDatatoPointData1 = paraview.simple.CellDatatoPointData(Input=shrink1)
    cellDatatoPointData1.CellDataArraytoprocess = t
    # Show data in view
    cell_data_to_point_data_display = paraview.simple.Show(
        cellDatatoPointData1, render_view, 'UnstructuredGridRepresentation'
    )
    # Trace defaults for the display properties.
    cell_data_to_point_data_display.Representation = 'Surface'
    cell_data_to_point_data_display.ColorArrayName = [None, '']
    cell_data_to_point_data_display.OSPRayScaleArray = \
        ['network | ' + network.name + ' | labels | pore.all']
    cell_data_to_point_data_display.OSPRayScaleFunction = 'PiecewiseFunction'
    cell_data_to_point_data_display.SelectOrientationVectors = 'None'
    cell_data_to_point_data_display.ScaleFactor = (maxshape-1)/10
    cell_data_to_point_data_display.SelectScaleArray = 'None'
    cell_data_to_point_data_display.GlyphType = 'Arrow'
    cell_data_to_point_data_display.GlyphTableIndexArray = 'None'
    cell_data_to_point_data_display.GaussianRadius = (maxshape-1)/200
    cell_data_to_point_data_display.SetScaleArray = \
        ['POINTS', 'network | ' + network.name + ' | labels | pore.all']
    cell_data_to_point_data_display.ScaleTransferFunction = 'PiecewiseFunction'
    cell_data_to_point_data_display.OpacityArray = \
        ['POINTS', 'network | ' + network.name + ' | labels | pore.all']
    cell_data_to_point_data_display.OpacityTransferFunction = 'PiecewiseFunction'
    cell_data_to_point_data_display.DataAxesGrid = 'GridAxesRepresentation'
    cell_data_to_point_data_display.PolarAxes = 'PolarAxesRepresentation'
    cell_data_to_point_data_display.ScalarOpacityUnitDistance = 1.0349360947089783
    # Init the 'PiecewiseFunction' selected for 'ScaleTransferFunction'
    cell_data_to_point_data_display.ScaleTransferFunction.Points = [
        1, 0, 0.5, 0, 1, 1, 0.5, 0]
    # Init the 'PiecewiseFunction' selected for 'OpacityTransferFunction'
    cell_data_to_point_data_display.OpacityTransferFunction.Points = [
        1, 0, 0.5, 0, 1, 1, 0.5, 0]
    # Hide data in view
    paraview.simple.Hide(shrink1, render_view)
    # Update the view to ensure updated data information
    render_view.Update()
    # Set active source
    paraview.simple.SetActiveSource(shrink1)
    # Set active source
    paraview.simple.SetActiveSource(cellDatatoPointData1)
    # Set active source
    paraview.simple.SetActiveSource(shrink1)
    # Create a new 'Extract Surface'
    extractSurface1 = paraview.simple.ExtractSurface(Input=shrink1)
    # Show data in view
    extract_surface_display = paraview.simple.Show(
        extractSurface1, render_view, 'GeometryRepresentation')
    # Trace defaults for the display properties.
    extract_surface_display.Representation = 'Surface'
    extract_surface_display.ColorArrayName = [None, '']
    extract_surface_display.OSPRayScaleArray = \
        ['network | ' + network.name + ' | labels | pore.all']
    extract_surface_display.OSPRayScaleFunction = 'PiecewiseFunction'
    extract_surface_display.SelectOrientationVectors = 'None'
    extract_surface_display.ScaleFactor = (maxshape-1)/10
    extract_surface_display.SelectScaleArray = 'None'
    extract_surface_display.GlyphType = 'Arrow'
    extract_surface_display.GlyphTableIndexArray = 'None'
    extract_surface_display.GaussianRadius = (maxshape-1)/200
    extract_surface_display.SetScaleArray = \
        ['POINTS', 'network | ' + network.name + ' | labels | pore.all']
    extract_surface_display.ScaleTransferFunction = 'PiecewiseFunction'
    extract_surface_display.OpacityArray = \
        ['POINTS', 'network | ' + network.name + ' | labels | pore.all']
    extract_surface_display.OpacityTransferFunction = 'PiecewiseFunction'
    extract_surface_display.DataAxesGrid = 'GridAxesRepresentation'
    extract_surface_display.PolarAxes = 'PolarAxesRepresentation'
    # Init the 'PiecewiseFunction' selected for 'ScaleTransferFunction'
    extract_surface_display.ScaleTransferFunction.Points = [
        1, 0, 0.5, 0, 1, 1, 0.5, 0]
    # Init the 'PiecewiseFunction' selected for 'OpacityTransferFunction'
    extract_surface_display.OpacityTransferFunction.Points = [
        1, 0, 0.5, 0, 1, 1, 0.5, 0]
    # Hide data in view
    paraview.simple.Hide(shrink1, render_view)
    # Update the view to ensure updated data information
    render_view.Update()
    # create a new 'Tube'
    tube = paraview.simple.Tube(Input=extractSurface1)
    tube.Scalars = ['POINTS', 'network |' + network.name + '| labels | pore.all']
    tube.Vectors = [None, '1']
    tube.Radius = 0.04
    # Set active source
    paraview.simple.SetActiveSource(extractSurface1)
    # Destroy tube
    paraview.simple.Delete(tube)
    del tube
    # Set active source
    paraview.simple.SetActiveSource(shrink1)
    # Set active source
    paraview.simple.SetActiveSource(cellDatatoPointData1)
    # Set active source
    paraview.simple.SetActiveSource(extractSurface1)
    # Create a new 'Tube'
    tube = paraview.simple.Tube(Input=extractSurface1)
    tube.Scalars = [
        'POINTS', 'network |' + network.name + ' | labels | pore.all']
    tube.Vectors = [None, '1']
    tube.Radius = 0.04
    # Properties modified on tube
    tube.Vectors = ['POINTS', '1']
    # Show data in view
    tube_display = paraview.simple.Show(tube, render_view,
                                        'GeometryRepresentation')
    # Trace defaults for the display properties.
    tube_display.Representation = 'Surface'
    tube_display.ColorArrayName = [None, '']
    tube_display.OSPRayScaleArray = 'TubeNormals'
    tube_display.OSPRayScaleFunction = 'PiecewiseFunction'
    tube_display.SelectOrientationVectors = 'None'
    tube_display.ScaleFactor = (maxshape)/10
    tube_display.SelectScaleArray = 'None'
    tube_display.GlyphType = 'Arrow'
    tube_display.GlyphTableIndexArray = 'None'
    tube_display.GaussianRadius = (maxshape)/200
    tube_display.SetScaleArray = ['POINTS', 'TubeNormals']
    tube_display.ScaleTransferFunction = 'PiecewiseFunction'
    tube_display.OpacityArray = ['POINTS', 'TubeNormals']
    tube_display.OpacityTransferFunction = 'PiecewiseFunction'
    tube_display.DataAxesGrid = 'GridAxesRepresentation'
    tube_display.PolarAxes = 'PolarAxesRepresentation'
    # Init the 'PiecewiseFunction' selected for 'ScaleTransferFunction'
    tube_display.ScaleTransferFunction.Points = [-1, 0, 0.5, 0, 1, 1, 0.5, 0]
    # Init the 'PiecewiseFunction' selected for 'OpacityTransferFunction'
    tube_display.OpacityTransferFunction.Points = [-1, 0, 0.5, 0, 1, 1, 0.5, 0]
    # Hide data in view
    paraview.simple.Hide(extractSurface1, render_view)
    # Update the view to ensure updated data information
    render_view.Update()
    # Saving camera placements for all active views
    # Current camera placement for render_view
    render_view.CameraPosition = [
        (xshape + 1) / 2, (yshape + 1) / 2, 4.3 * np.sqrt(np.sum(shape / 2)**2)
    ]
    render_view.CameraFocalPoint = [(xi+1) / 2 for xi in shape]
    render_view.CameraParallelScale = np.sqrt(np.sum(shape / 2)**2)
    paraview.simple.SaveState(f"{file}.pvsm")



import numpy as np
import scipy.sparse as sprs
from openpnm._skgraph import settings, tools


__all__ = [
    'trim_nodes',
    'trim_edges',
    'drop_nodes_from_am',
    'add_nodes',
    'add_edges',
    'split_edges',
]


def add_nodes(network, new_coords):
    r"""
    Given a list of node coordinates, add them to the network

    Parameters
    ----------
    network : dict
        A dictionary containing the node and edge attributes as ndarrays
    new_coords : ndarray
        The N-by-3 array of coordinates of the new nodes

    Returns
    -------
    network : dict
        The network dictionary with the new nodes added to the end. Note that
        any existing node attributes are also extended and filled with default
        values specified in ``settings.default_values``

    """

    # TODO: At present this does not work with empty networks.  It's a bit
    # challenging to determine what size each 'empty' array should become.
    # It's fine for 1D arrays like ``np.array([])``, but for N-dimensional
    # stuff it's trickier. For instance ``np.array([[], []])`` has a shape
    # of (2, 0) so the empty dimension is the wrong one since all the array
    # extending in this function occurs on the 1st axis.
    g = network
    node_prefix = tools.get_node_prefix(g)
    coords = np.atleast_2d(new_coords)
    Nnew = coords.shape[0]
    for k, v in g.items():
        if k.startswith(node_prefix):
            dval = None
            for t in settings.missing_values.keys():
                if v.dtype == t:
                    dval = settings.missing_values[t]
            blank = np.repeat(v[:1, ...], Nnew, axis=0)*dval
            blank.fill(dval)
            g[k] = np.concatenate((v, blank), axis=0).astype(v.dtype)
    # Lastly, overwrite the -Nnew elements of coords with the given values
    g[node_prefix+'.coords'][-Nnew:] = np.array(coords)
    return g


def add_edges(network, new_conns):
    r"""
    Given a list of edge connections, add them to the network

    Parameters
    ----------
    network : dict
        A dictionary containing the node and edge attributes as ndarrays
    new_conns : ndarray
        The N-by-2 array of connections betwween existing nodes

    Returns
    -------
    network : dict
        The network dictionary with the new edges added to the end. Note that
        any existing edge attributes are also extended and filled with default
        values specified in ``settings.default_values``.

    """

    # TODO: At present this does not work with empty networks.  It's a bit
    # challenging to determine what size each 'empty' array should become.
    # It's fine for 1D arrays like ``np.array([])``, but for N-dimensional
    # stuff it's trickier. For instance ``np.array([[], []])`` has a shape
    # of (2, 0) so the empty dimension is the wrong one since all the array
    # extending in this function occurs on the 1st axis.
    g = network
    edge_prefix = tools.get_edge_prefix(g)
    conns = np.atleast_2d(new_conns)
    Nnew = conns.shape[0]
    for k, v in g.items():
        if k.startswith(edge_prefix):
            dval = None
            for t in settings.missing_values.keys():
                if v.dtype == t:
                    dval = settings.missing_values[t]
            blank = np.repeat(v[:1, ...], Nnew, axis=0)*dval
            blank.fill(dval)
            g[k] = np.concatenate((v, blank), axis=0).astype(v.dtype)
    # Lastly, overwrite the -Nnew elements of coords with the given values
    g[edge_prefix+'.conns'][-Nnew:] = np.array(conns)
    return g


def trim_edges(network, inds):
    r"""
    Removes given edges from a graph or network

    Parameters
    ----------
    network : dictionary
        A dictionary containing coords, conns and other attributes
    inds : array_like
        The edge indices to be trimmed in the form of a 1D list or boolean
        mask with ``True`` values indicating indices to trim.

    Returns
    -------
    network : dict
        The dictionary with all edge arrays trimmed accordingly

    """
    g = network
    edge_prefix = tools.get_edge_prefix(g)
    N_bonds = g[edge_prefix+'.conns'].shape[0]
    inds = np.atleast_1d(inds)
    keep = np.ones(N_bonds, dtype=bool)
    keep[inds] = False
    for k, v in g.items():
        if k.startswith(edge_prefix):
            g[k] = v[keep]
    return g


def trim_nodes(network, inds):
    r"""
    Removes given nodes and any connected edges from a graph or network

    Parameters
    ----------
    network : dict
        A dictionary containing coords, conns and other attributes
    inds : array_like
        The node indices to be trimmed in the form of a 1D list or boolean
        mask with ``True`` values indicating indices to trim.

    Returns
    -------
    network : dict
        The dictionary with all nodes arrays trimmed accordingly, all edges
        trimmed that were connected to the trimmed nodes, and the 'edge.conns'
        array renumbered so edges point to the updated node indices.

    """
    node_prefix = tools.get_node_prefix(network)
    edge_prefix = tools.get_edge_prefix(network)
    N_sites = network[node_prefix+'.coords'].shape[0]
    inds = np.atleast_1d(inds)
    if inds.dtype == bool:
        inds = np.where(inds)[0]
    keep = np.ones(N_sites, dtype=bool)
    keep[inds] = False
    for k, v in network.items():
        if k.startswith(node_prefix):
            network[k] = v[keep]
    # Remove edges
    edges = np.any(np.isin(network[edge_prefix+'.conns'], inds), axis=1)
    network = trim_edges(network, inds=edges)
    # Renumber conns
    remapping = np.cumsum(keep) - 1
    network[edge_prefix+'.conns'] = remapping[network[edge_prefix+'.conns']]
    return network


def drop_nodes_from_am(am, inds):
    r"""
    Update adjacency matrix after dropping nodes

    Parameters
    ----------
    am : scipy.sparse matrix
        The adjacency matrix of the network in COO format.
    inds : array_like
        A list of which nodes indices to drop.  Can either be integer indices
        or a boolean mask with ``True`` indicating which locations to drop

    Returns
    -------
    am : ndarray
        An updated adjacency matrix with nodes and headless edges removed,
        and node indices updated accordingly
    dropped_edges : ndarray
        A boolean array with ``True`` values indicating which edges
        were rendered headless. This can be used to drop invalid edges
        from other arrays (i.e. array = array[~dropped_edges]).

    """
    nodes = np.array(inds)
    if nodes.dtype != bool:
        inds = np.copy(nodes)
        nodes = np.zeros(am.shape[0], dtype=bool)
        nodes[inds] = True
    node_mask = ~nodes
    conns = np.vstack((am.row, am.col)).T
    node_id = np.cumsum(node_mask) - 1
    edge_mask = ~np.all(node_mask[conns], axis=1)
    conns = node_id[conns[~edge_mask]]
    am = sprs.coo_matrix((am.data[~edge_mask], (conns[:, 0], conns[:, 1])))
    return am, edge_mask


def split_edges(network):
    r"""
    Inserts an new node between each existing node and joins with new edges

    Parameters
    ----------
    network : dict
        The dictionary containing the network connections and coordinates

    Returns
    -------
    result : tuple
        A tuple containing ``new_conns`` and optionally ``new_coords`` if
        ``coords`` was provided.

        ============== ========================================================
        Value          Description
        ============== ========================================================
        ``new_conns``  A new adjacency matrix in COO format with new nodes
                       added between each original node. If edge 1 connected
                       nodes 1 and 2, then row 1 of the new sparse adjacency
                       matrix will be [1, Nt + 1], and row Nt + 1 will be
                       [Nt + 1, 2].
        ``new_coords`` A and updated list of node coordinates with the new
                       nodes appended to the end.  The coordinates of the new
                       nodes are taken as the average of the two nodes between
                       which they were inserted.
        ============== ========================================================

    """
    g = network
    node_prefix = tools.get_node_prefix(g)
    edge_prefix = tools.get_edge_prefix(g)
    conns = g[edge_prefix + '.conns']
    coords = g[node_prefix + '.coords']
    Nt = conns.shape[0]
    Np = conns.max() + 1
    new_conns = np.zeros([2*Nt, 2], dtype=int)
    new_conns[:Nt, :] = np.vstack((conns[:, 0], np.arange(Np, Np+Nt))).T
    new_conns[Nt:, :] = np.vstack((np.arange(Np, Np+Nt), conns[:, 1])).T
    result = (new_conns, )
    if coords is not None:
        Np = coords.shape[0]
        new_coords = np.zeros([Np + Nt, 3], dtype=float)
        new_coords[:Np, :] = coords
        new_coords[Np:, :] = np.mean(coords[conns], axis=1)
        result = (new_conns, new_coords)
    return result



import numpy as np
from openpnm.models.phase import _phasedocs


__all__ = [
    'water_correlation',
    'gas_pure_gismr',
    'liquid_pure_gismr',
    'liquid_pure_sr',
    'gas_mixture_whz',
    'liquid_mixture_DIPPR9H',
]


@_phasedocs
def water_correlation(
    phase,
    T="pore.temperature",
    salinity="pore.salinity",
):
    r"""
    Calculates thermal conductivity of pure water or seawater at atmospheric
    pressure using the correlation given [1].

    Values at temperature higher the normal boiling temperature are calculated
    at the saturation pressure.

    Parameters
    ----------
    %(phase)s
    %(T)s
    %(salinity)s

    Returns
    -------
    value : ndarray
        A numpy ndarray containing thermal conductivity of water/seawater in [W/m.K]

    Notes
    -----
    T must be in K, and S in g of salt per kg of phase, or ppt (parts per
    thousand). The correlation is valid for 273 < T < 453 K and
    0 < S < 160 g/kg within 3 percent accuracy.

    References
    ----------
    [1] D. T. Jamieson, and J. S. Tudhope, Desalination, 8, 393-401, 1970.

    """
    T = phase[T]
    if salinity in phase.keys():
        S = phase[salinity]
    else:
        S = 0
    T68 = 1.00024 * T  # convert from T_90 to T_68
    SP = S / 1.00472  # convert from S to S_P
    k_sw = 0.001 * (
        10
        ** (
            np.log10(240 + 0.0002 * SP)
            + 0.434
            * (2.3 - (343.5 + 0.037 * SP) / T68)
            * ((1 - T68 / (647.3 + 0.03 * SP))) ** (1 / 3)
        )
    )
    value = k_sw
    return value


@_phasedocs
def gas_pure_gismr(
    phase,
    T='pore.temperature',
    MW='param.molecular_weight',
    Tb='param.boiling_temperature',
    Pc='param.critical_pressure',
    omega='param.acentric_factor',
):
    r"""
    Calculate the thermal conductivty of a pure gas using the correlation in
    [1]

    Parameters
    ----------
    %(phase)s
    %(T)s
    %(MW)s
    %(Tb)s
    %(Pc)s
    %(omega)s

    Returns
    -------

    References
    ----------
    [1] gharagheizi method: doi:10.1002/aic.13938

    """
    T = phase[T]
    MW = phase[MW]
    Tb = phase[Tb]
    # The following correction suggested by chemicals package author
    Pc = phase[Pc]/10000
    omega = phase[omega]
    B = (T + (2.0*omega + 2.0*T - 2.0*T*(2.0*omega + 3.2825)/Tb + 3.2825)
         / (2.0*omega + T - T*(2.0*omega + 3.2825)/Tb + 3.2825)
         - T*(2.0*omega + 3.2825)/Tb)
    A = (2*omega + T - (2*omega + 3.2825)*T/Tb + 3.2825)/(0.1*MW*Pc*T) \
        * (3.9752*omega + 0.1*Pc + 1.9876*B + 6.5243)**2
    k = 7.9505e-4 + 3.989e-5*T - 5.419e-5*MW + 3.989e-5*A
    return k


@_phasedocs
def liquid_pure_gismr(
    phase,
    T='pore.temperature',
    MW='param.molecular_weight',
    Tb='param.boiling_temperature',
    Pc='param.critical_pressure',
    omega='param.acentric_factor',
):
    r"""
    Calculates the thermal conductivity of a pure liquid using the correlation
    in [1]

    Parameters
    ----------
    %(phase)s
    %(T)s
    %(MW)s
    %(Tb)s
    %(Pc)s
    %(omega)s

    Returns
    -------

    References
    ----------
    [1] gharagheizi method: doi:10.1002/aic.13938

    """
    T = phase[T]
    MW = phase[MW]
    Tb = phase[Tb]
    Pc = phase[Pc]/100000
    omega = phase[omega]
    B = 16.0407*MW + 2.0*Tb - 27.9074
    A = 3.8588*(MW**8)*(1.0045*B + 6.5152*MW - 8.9756)
    k = (1e-4)*(10*omega + 2*Pc - 2*T + 4 + 1.908*(Tb + 1.009*(B**2)/(MW**2))
                + 3.9287*(MW**4)/(B**4) + A/(B**8))
    return k


@_phasedocs
def liquid_pure_sr(
    phase,
    T="pore.temperature",
    Tc='param.critical_temperature',
    MW='param.molecular_weight',
    Tb='param.boiling_temperature',
):
    r"""
    Calculates the thermal conductivity of a pure liquid using the correlation
    in [1]

    Parameters
    ----------
    %(phase)s
    %(T)s
    %(Tc)s
    %(MW)s
    %(Tb)s

    Returns
    -------
    value : ndarray
        A numpy ndarray containing thermal conductivity values in [W/m.K]

    References
    ----------
    [1] Sato et al

    """
    T = phase[T]
    Tc = phase[Tc]
    MW = phase[MW]
    Tbr = phase[Tb]/Tc
    Tr = T / Tc
    value = ((1.1053 / ((MW) ** 0.5)) * (3 + 20 * (1 - Tr) ** (2 / 3))
             / (3 + 20 * (1 - Tbr) ** (2 / 3)))
    return value


@_phasedocs
def liquid_mixture_DIPPR9I(
    phase,
    rhos='pore.density.*',
    ks='pore.thermal_conductivity.*',
    MWs='param.molecular_weight.*',
):
    r"""
    Calculates the thermal conductivity of a liquid mixture using the
    correlation in [1]

    Parameters
    ----------
    %(phase)s
    %(rhos)s
    %(ks)s
    %(MWs)s

    Returns
    -------

    References
    ----------
    [1] DIPPR9I

    """
    raise NotImplementedError("This function is not ready yet")
    from chemicals import rho_to_Vm
    xs = phase['pore.mole_fraction']
    kLs = phase.get_comp_vals(ks)
    # kL = numba_vectorized.DIPPR9I(xs, kLs)  # Another one that doesn't work
    Vm = [rho_to_Vm(c.get_comp_vals(rhos), c.get_comp_vals(MWs))
          for c in phase.components.keys()]
    denom = np.sum([xs[i]*Vm[i] for i in range(len(xs))], axis=0)
    phis = np.array([xs[i]*Vm[i] for i in range(len(xs))])/denom
    kij = 2/np.sum([1/kLs[i] for i in range(len(xs))], axis=0)
    kmix = np.zeros_like(xs[0])
    N = len(xs)
    for i in range(N):
        for j in range(N):
            kmix += phis[i]*phis[j]*kij
    return kmix


@_phasedocs
def liquid_mixture_DIPPR9H(
    phase,
    ks='pore.thermal_conductivity.*',
    MWs='param.molecular_weight.*',
):
    r"""
    Calculates the thermal conductivity of a liquid mixture using the
    correlation in [1]

    Parameters
    ----------
    %(phase)s
    %(ks)s
    %(MWs)s

    Returns
    -------

    References
    ----------
    [1] DIPPR9H

    """
    xs = phase['pore.mole_fraction']
    MW = phase.get_comp_vals(MWs)
    ks = phase.get_comp_vals(ks)
    num = np.vstack([xs[k]*MW[k]/(ks[k]**2) for k in xs.keys()]).sum(axis=0)
    denom = np.vstack([xs[k]*MW[k] for k in xs.keys()]).sum(axis=0)
    temp = num/denom
    kmix = (1/temp)**0.5
    return kmix


@_phasedocs
def gas_mixture_whz(
    phase,
    T='pore.temperature',
    ks='pore.thermal_conductivity.*',
    MWs='param.molecular_weight.*',
):
    r"""
    Calculates the viscosity of a gas mixture using the correlation in [1]

    Parameters
    ----------
    %(phase)s
    %(T)s
    %(ks)s
    %(MWs)s

    Returns
    -------

    References
    ----------
    [1] Wassiljew, Herning & Zipperer
    """
    T = phase[T]
    ys = phase['pore.mole_fraction']
    kGs = phase.get_comp_vals(ks)
    MWs = phase.get_comp_vals(MWs)
    kmix = np.zeros_like(T)
    for i, ki in enumerate(ys.keys()):
        num = ys[ki]*kGs[ki]
        denom = 0.0
        for j, kj in enumerate(ys.keys()):
            A = np.sqrt(MWs[kj]/MWs[ki])
            denom += ys[ki]*A
        kmix += num/denom
    return kmix



import logging
from openpnm.algorithms import ReactiveTransport
from openpnm.utils import Docorator


docstr = Docorator()
logger = logging.getLogger(__name__)


__all__ = ['OhmicConduction']


@docstr.dedent
class OhmicConductionSettings:
    r"""

    Parameters
    ----------
    %(ReactiveTransportSettings.parameters)s

    """
    quantity = 'pore.voltage'
    conductance = 'throat.electrical_conductance'


class OhmicConduction(ReactiveTransport):
    r"""
    A subclass of LinearTransport to simulate electron and ionic
    conduction.  The 2 main roles of this subclass are to set the default
    property names and to implement a method for calculating the effective
    conductivity of the network.

    """

    def __init__(self, name='ohmic_?', **kwargs):
        super().__init__(name=name, **kwargs)
        self.settings._update(OhmicConductionSettings())



import numpy as np


def cubic(shape, spacing=1, connectivity=6, node_prefix='node', edge_prefix='edge'):
    r"""
    Generate a simple cubic lattice

    Parameters
    ----------
    shape : array_like
        The number of unit cells in each direction.  A unit cell has 1 vertex
        at its center.
    spacing : array_like or float
        The size of a unit cell in each direction. If an scalar is given it is
        applied in all 3 directions.

    Returns
    -------
    network : dict
        A dictionary containing ``coords`` and ``conns`` of a cubic network with the
        specified spacing and connectivity.

    """
    # Take care of 1D/2D networks
    shape = np.array(shape, ndmin=1)
    shape = np.concatenate((shape, [1] * (3 - shape.size))).astype(int)
    arr = np.atleast_3d(np.empty(shape))
    spacing = np.float64(spacing)
    if spacing.size == 2:
        spacing = np.concatenate((spacing, [1]))
    spacing = np.ones(3, dtype=float) * np.array(spacing, ndmin=1)

    z = np.tile(np.arange(shape[2]), shape[0] * shape[1])
    y = np.tile(np.repeat(np.arange(shape[1]), shape[2]), shape[0])
    x = np.repeat(np.arange(shape[0]), shape[1] * shape[2])
    points = (np.vstack([x, y, z]).T).astype(float) + 0.5

    idx = np.arange(arr.size).reshape(arr.shape)

    face_joints = [(idx[:, :, :-1], idx[:, :, 1:]),
                   (idx[:, :-1], idx[:, 1:]),
                   (idx[:-1], idx[1:])]

    corner_joints = [(idx[:-1, :-1, :-1], idx[1:, 1:, 1:]),
                     (idx[:-1, :-1, 1:], idx[1:, 1:, :-1]),
                     (idx[:-1, 1:, :-1], idx[1:, :-1, 1:]),
                     (idx[1:, :-1, :-1], idx[:-1, 1:, 1:])]

    edge_joints = [(idx[:, :-1, :-1], idx[:, 1:, 1:]),
                   (idx[:, :-1, 1:], idx[:, 1:, :-1]),
                   (idx[:-1, :, :-1], idx[1:, :, 1:]),
                   (idx[1:, :, :-1], idx[:-1, :, 1:]),
                   (idx[1:, 1:, :], idx[:-1, :-1, :]),
                   (idx[1:, :-1, :], idx[:-1, 1:, :])]

    if connectivity == 6:
        joints = face_joints
    elif connectivity == 6 + 8:
        joints = face_joints + corner_joints
    elif connectivity == 6 + 12:
        joints = face_joints + edge_joints
    elif connectivity == 12 + 8:
        joints = edge_joints + corner_joints
    elif connectivity == 6 + 8 + 12:
        joints = face_joints + corner_joints + edge_joints
    else:
        raise Exception("Invalid connectivity. Must be 6, 14, 18, 20 or 26.")

    tails, heads = np.array([], dtype=int), np.array([], dtype=int)
    for T, H in joints:
        tails = np.concatenate((tails, T.flatten()))
        heads = np.concatenate((heads, H.flatten()))
    pairs = np.vstack([tails, heads]).T
    # NOTE: pairs is already sorted for connectivity = 6
    if connectivity != 6:
        pairs = np.sort(pairs, axis=1)

    d = {}
    d[f"{node_prefix}.coords"] = points * spacing
    d[f"{edge_prefix}.conns"] = pairs

    return d



import logging
import numpy as _np
from openpnm.models import misc as _misc
from openpnm.models.geometry import _geodocs


__all__ = [
    "weibull",
    "normal",
    "random",
    "generic_distribution",
    "from_neighbor_throats",
    "largest_sphere",
    "equivalent_diameter"
]


logger = logging.getLogger(__name__)


def weibull(network, shape, scale, loc, seeds='pore.seed'):
    return _misc.weibull(network, shape=shape, scale=scale, loc=loc,
                         seeds=seeds)


weibull.__doc__ = _misc.weibull.__doc__


def normal(network, scale, loc, seeds='pore.seed'):
    return _misc.normal(network, scale=scale, loc=loc,
                        seeds=seeds)


normal.__doc__ = _misc.normal.__doc__


def random(network, seed=None, num_range=[0, 1]):
    return _misc.random(network, element='pore', seed=seed,
                        num_range=num_range)


random.__doc__ = _misc.random.__doc__


def generic_distribution(network, func, seeds='pore.seed'):
    return _misc.generic_distribution(network, func=func, seeds=seeds)


generic_distribution.__doc__ = _misc.generic_distribution.__doc__


def from_neighbor_throats(network, prop, mode='max'):
    return _misc.from_neighbor_throats(network,
                                       prop=prop,
                                       mode=mode)


from_neighbor_throats.__doc__ = _misc.from_neighbor_throats.__doc__


@_geodocs
def largest_sphere(
    network,
    fixed_diameter='pore.fixed_diameter',
    iters=5
):
    r"""
    Finds the maximum diameter pore that can be placed in each location without
    overlapping any neighbors.

    Parameters
    ----------
    %(network)s
    fixed_diameter : str
        Name of the dictionary key on ``network`` where the array containing
        pore diameter values is stored, if any. If not provided a starting
        value is assumed as half-way to the nearest neighbor.
    iters : integer
        The number of iterations to perform when searching for maximum
        diameter.  This function iteratively grows pores until they touch
        their nearest neighbor, which is also growing, so this parameter limits
        the maximum number of iterations.  The default is 10.

    Returns
    -------
    diameters : ndarray
        A numpy ndarray containing pore diameter values

    Notes
    -----
    This function iteratively expands pores by increasing their diameter to
    encompass half of the distance to the nearest neighbor.  If the neighbor
    is not growing because it's already touching a different neighbor, then
    the given pore will never quite touch this neighbor.  Increasing the value
    of ``iters`` will get it closer, but it's case of Zeno's paradox with
    each step cutting the remaining distance in half.

    This model looks into all pores in the network when finding the diameter.
    This means that when multiple Geometry objects are defined, it will
    consider the diameter of pores on adjacent Geometries. If no diameters
    have been assigned to these neighboring pores it will assume 0. If
    diameter value are assigned to the neighboring pores AFTER this model is
    run, the pores may overlap.  This can be remedied by running this model
    again.

    """
    P12 = network['throat.conns']
    C1 = network['pore.coords'][network['throat.conns'][:, 0]]
    C2 = network['pore.coords'][network['throat.conns'][:, 1]]
    L = _np.sqrt(_np.sum((C1 - C2)**2, axis=1))
    try:
        # Fetch any existing pore diameters on the network
        D = network[fixed_diameter]
        # Set any unassigned values (nans) to 0
        D[_np.isnan(D)] = 0
    except KeyError:
        logger.info('Pore sizes not present, calculating starting values '
                    + 'as half-way to the nearest neighbor')
        D = _np.inf*_np.ones([network.Np, ], dtype=float)
        _np.minimum.at(D, P12[:, 0], L)
        _np.minimum.at(D, P12[:, 1], L)
    while iters >= 0:
        iters -= 1
        Lt = L - _np.sum(D[P12], axis=1)/2
        Dadd = _np.ones_like(D)*_np.inf
        _np.minimum.at(Dadd, P12[:, 0], Lt)
        _np.minimum.at(Dadd, P12[:, 1], Lt)
        D += Dadd
    if _np.any(D < 0):
        logger.info('Negative pore diameters found!  Neighboring pores are '
                    + 'larger than the pore spacing.')
    return D


@_geodocs
def equivalent_diameter(
    network,
    pore_volume='pore.volume',
    pore_shape='sphere'
):
    r"""
    Calculate the diameter of a sphere or edge-length of a cube with same
    volume as the pore.

    Parameters
    ----------
    %(network)s
    %(Vp)s
    pore_shape : str
        The shape of the pore body to assume when back-calculating from
        volume.  Options are 'sphere' (default) or 'cube'.

    Returns
    -------
    diameters : ndarray
        A number ndarray containing pore diameter values

    """
    from scipy.special import cbrt
    pore_vols = network[pore_volume]
    if pore_shape.startswith('sph'):
        value = cbrt(6*pore_vols/_np.pi)
    elif pore_shape.startswith('cub'):
        value = cbrt(pore_vols)
    return value



import logging
import numpy as np
from openpnm.io import _parse_filename


logger = logging.getLogger(__name__)


_header = """
import numpy as np
def pnm_2_salome(cylinder_head, cylinder_tail, cylinder_r,
                 sphere_c, sphere_r, explicit=False):
    import numpy as np
    import math
    import GEOM
    import SALOMEDS
    import salome
    from salome.geom import geomBuilder
    salome.salome_init()
    geompy = geomBuilder.New()
    O = geompy.MakeVertex(0, 0, 0)
    OX = geompy.MakeVectorDXDYDZ(1, 0, 0)
    OY = geompy.MakeVectorDXDYDZ(0, 1, 0)
    OZ = geompy.MakeVectorDXDYDZ(0, 0, 1)
    s = len(sphere_r)
    c = len(cylinder_r)
    vertex_id = 0
    fuse_list = np.array([])
    for i in range(s):
        vertex_s = geompy.MakeVertex(sphere_c[i, 0], sphere_c[i, 1],
                                     sphere_c[i, 2])
        sphere = geompy.MakeSpherePntR(vertex_s, sphere_r[i])
        if explicit:
            geompy.addToStudy(vertex_s, "Vertex_{}".format(vertex_id))
            geompy.addToStudy(sphere, "Sphere_{}".format(i))
        vertex_id += 1
        fuse_list = np.append(fuse_list, sphere)
    for i in range(c):
        vertex_c1 = geompy.MakeVertex(cylinder_head[i, 0], cylinder_head[i, 1],
                                      cylinder_head[i, 2])
        vertex_c2 = geompy.MakeVertex(cylinder_tail[i, 0], cylinder_tail[i, 1],
                                      cylinder_tail[i, 2])
        line = geompy.MakeLineTwoPnt(vertex_c1, vertex_c2)
        length = abs(np.linalg.norm(cylinder_head[i] - cylinder_tail[i]))
        cylinder = geompy.MakeCylinder(vertex_c1, line, cylinder_r[i], length)
        if explicit:
            geompy.addToStudy(vertex_c1, "Vertex_{}".format(vertex_id))
        vertex_id += 1
        if explicit:
            geompy.addToStudy(vertex_c2, "Vertex_{}".format(vertex_id))
        vertex_id += 1
        if explicit:
            geompy.addToStudy(line, "Line_{}".format(i))
            geompy.addToStudy(cylinder, "Cylinder_{}".format(i))
        fuse_list = np.append(fuse_list, cylinder)
    fuse = geompy.MakeFuseList(list(fuse_list), True, True)
    geompy.addToStudy(fuse, 'network')
    geompy.addToStudy( O, 'O' )
    geompy.addToStudy( OX, 'OX' )
    geompy.addToStudy( OY, 'OY' )
    geompy.addToStudy( OZ, 'OZ' )
    if salome.sg.hasDesktop():
        salome.sg.updateObjBrowser()

    """
_footer = """
pnm_2_salome(cylinder_head, cylinder_tail, cylinder_r,
             sphere_c, sphere_r, explicit=explicit)

    """


def network_to_salome(network, filename=None, explicit=False):
    r"""
    Saves the network data and writes a Salome .py instruction file.

    Parameters
    ----------
    network : Network
        The network containing the desired data

    Notes
    -----
    This method only saves the data, not any of the pore-scale models
    or other attributes. To save an actual OpenPNM Project use the
    ``Workspace`` object.

    """
    # Temporarily add endpoints to network so STL class works
    network["throat.endpoints.head"] = network.coords[network.conns[:, 0]]
    network["throat.endpoints.tail"] = network.coords[network.conns[:, 1]]

    filename = network.name if filename is None else filename
    filename = _parse_filename(filename=filename, ext='py')

    f = open(filename, 'w')
    f.write(_header+'\n')

    x = network['throat.endpoints.head']
    s = str(x.shape)
    f.write('cylinder_head = np.array([')
    np.savetxt(f, X=[x.flatten()], fmt='%.18e', delimiter=',', newline='')
    f.write('])\n')
    f.write('cylinder_head = np.reshape(cylinder_head, '+s+')\n')

    x = network['throat.endpoints.tail']
    s = str(x.shape)
    f.write('cylinder_tail = np.array([')
    np.savetxt(f, X=[x.flatten()], fmt='%.18e', delimiter=',', newline='')
    f.write('])\n')
    f.write('cylinder_tail = np.reshape(cylinder_tail, '+s+')\n')

    x = network['throat.diameter']/2
    s = str(x.shape)
    f.write('cylinder_r = np.array([')
    np.savetxt(f, X=[x.flatten()], fmt='%.18e', delimiter=',', newline='')
    f.write('])\n')
    f.write('cylinder_r = np.reshape(cylinder_r, '+s+')\n')

    x = network['pore.coords']
    s = str(x.shape)
    f.write('sphere_c = np.array([')
    np.savetxt(f, X=[x.flatten()], fmt='%.18e', delimiter=',', newline='')
    f.write('])\n')
    f.write('sphere_c = np.reshape(sphere_c, '+s+')\n')

    x = network['pore.diameter']/2
    s = str(x.shape)
    f.write('sphere_r = np.array([')
    np.savetxt(f, X=[x.flatten()], fmt='%.18e', delimiter=',', newline='')
    f.write('])\n')
    f.write('sphere_r = np.reshape(sphere_r, '+s+')\n')

    f.write('explicit = '+str(explicit))

    f.write(_footer)

    f.close()



import logging
import sys
import numpy as np
from numpy.linalg import norm
try:  # For scipy < 1.14
    from scipy.optimize.nonlin import TerminationCondition
except ImportError:  # For newer Scipy
    from scipy.optimize._nonlin import TerminationCondition
from tqdm.auto import tqdm
from openpnm.algorithms import Transport
from openpnm.utils import Docorator, TypedList


__all__ = ["ReactiveTransport"]


docstr = Docorator()
logger = logging.getLogger(__name__)


@docstr.get_sections(base="ReactiveTransportSettings", sections=["Parameters"])
@docstr.dedent
class ReactiveTransportSettings:
    r"""

    Parameters
    ----------
    %(TransportSettings.parameters)s
    sources : list
        List of source terms that have been added
    relaxation_factor : float (default = 1.0)
        A relaxation factor to control under-relaxation for the quantity
        solving for. Factor approaching 0 leads to improved stability but
        slower simulation. Factor approaching 1 gives fast simulation but
        may be unstable.
    newton_maxiter : int
        Maximum number of iterations allowed for the nonlinear solver to
        converge.
    f_rtol : float
        Relative tolerance for the solution residual
    x_rtol : float
        Relative tolerance for the solution vector

    """
    relaxation_factor = 1.0
    newton_maxiter = 5000
    f_rtol = 1e-6
    x_rtol = 1e-6


@docstr.get_sections(base="ReactiveTransport", sections=["Parameters"])
@docstr.dedent
class ReactiveTransport(Transport):
    r"""
    A subclass for steady-state simulations with (optional) source terms.

    Parameters
    ----------
    %(Transport.parameters)s

    Notes
    -----
    This subclass performs steady simulations of transport phenomena with
    reactions when source terms are added.

    """

    def __init__(self, name="react_trans_?", **kwargs):
        super().__init__(name=name, **kwargs)
        self.settings._update(ReactiveTransportSettings())

    def set_source(self, pores, propname, mode="add"):
        r"""
        Applies a given source term to the specified pores

        Parameters
        ----------
        pores : array_like
            The pore indices where the source term should be applied.
        propname : str
            The property name of the source term model to be applied.
        mode : str
            Controls how the sources are applied. Options are:

            ===========  =====================================================
            mode         meaning
            ===========  =====================================================
            'add'        (default) Adds supplied source term to already
                         existing ones.
            'remove'     Deletes given source term from the specified
                         locations.
            'clear'      Deletes given source term from all locations (ignores
                         the ``pores`` argument).
            ===========  =====================================================

        Notes
        -----
        Source terms cannot be applied in pores where boundary conditions
        have already been set. Attempting to do so will result in an error
        being raised.

        """
        # If a list of modes was sent, do them each in order
        if isinstance(mode, list):
            for item in mode:
                self.set_source(pores=pores, propname=propname, mode=item)
            return
        propname = self._parse_prop(propname, "pore")
        # Check if any BC is already set in the same locations
        locs_BC = np.zeros(self.Np, dtype=bool)
        for item in self["pore.bc"].keys():
            locs_BC = np.isfinite(self[f"pore.bc.{item}"])
        if np.any(locs_BC[pores]):
            raise Exception("BCs present in given pores, can't assign source term")
        prop = "pore.source." + propname.split(".", 1)[1]
        if mode == "add":
            if prop not in self.keys():
                self[prop] = False
            self[prop][pores] = True
        elif mode == "remove":
            self[prop][pores] = False
        elif mode == "clear":
            self[prop] = False
        else:
            raise Exception(f"Unsupported mode {mode}")

    def _apply_sources(self):
        """
        Updates ``A`` and ``b``, applying source terms to specified pores.

        Notes
        -----
        Phase and physics objects are also updated before applying source
        terms to ensure that source terms values are associated with the
        current value of 'quantity'.

        """
        try:
            phase = self.project[self.settings.phase]
            for item in self["pore.source"].keys():
                # Fetch linearized values of the source term
                Ps = self["pore.source." + item]
                S1, S2 = [phase[f"pore.{item}.{Si}"] for Si in ["S1", "S2"]]
                # Modify A and b: diag(A) += -S1, b += S2
                diag = self.A.diagonal()
                diag[Ps] += -S1[Ps]
                self.A.setdiag(diag)
                self.b[Ps] += S2[Ps]
        except KeyError:
            pass

    def _run_special(self, solver, x0, verbose=None):
        r"""
        Repeatedly updates ``A``, ``b``, and the solution guess within
        according to the applied source term then calls ``_solve`` to
        solve the resulting system of linear equations.

        Stops when the max-norm of the residual drops by at least
        ``f_rtol``:

            ``norm(R_n) < norm(R_0) * f_rtol``

        AND

            ``norm(dx) < norm(x) * x_rtol``

        where R_i is the residual at ith iteration, x is the solution at
        current iteration, and dx is the change in the solution between two
        consecutive iterations. ``f_rtol`` and ``x_rtol`` are defined in
        the algorithm's settings under: ``alg.settings['f_rtol']``, and
        ``alg.settings['x_rtol']``, respectively.

        Parameters
        ----------
        x0 : ndarray
            Initial guess of the unknown variable

        """
        w = self.settings["relaxation_factor"]
        maxiter = self.settings["newton_maxiter"]
        f_rtol = self.settings["f_rtol"]
        x_rtol = self.settings["x_rtol"]
        xold = self.x
        dx = self.x - xold
        condition = TerminationCondition(
            f_tol=np.inf, f_rtol=f_rtol, x_rtol=x_rtol, norm=norm
        )

        tqdm_settings = {
            "total": 100,
            "desc": f"{self.name} : Newton iterations",
            "disable": not verbose,
            "file": sys.stdout,
            "leave": False,
        }

        with tqdm(**tqdm_settings) as pbar:
            for i in range(maxiter):
                res = self._get_residual()
                progress = self._get_progress(res)
                pbar.update(progress - pbar.n)
                is_converged = bool(condition.check(f=res, x=xold, dx=dx))
                if is_converged:
                    pbar.update(100 - pbar.n)
                    self.soln.is_converged = is_converged
                    logger.info(f"Solution converged, residual norm: {norm(res):.4e}")
                    return
                super()._run_special(solver=solver, x0=xold, w=w)
                dx = self.x - xold
                xold = self.x
                logger.info(f"Iteration #{i:<4d} | Residual norm: {norm(res):.4e}")
                self.soln.num_iter = i + 1

        self.soln.is_converged = False
        logger.warning(f"{self.name} didn't converge after {maxiter} iterations")

    def _get_progress(self, res):
        """
        Returns an approximate value for completion percent of Newton iterations.
        """
        if not hasattr(self, "_f0_norm"):
            self._f0_norm = norm(res)
        f_rtol = self.settings.f_rtol
        norm_reduction = norm(res) / self._f0_norm / f_rtol
        progress = (1 - max(np.log10(norm_reduction), 0) / np.log10(1 / f_rtol)) * 100
        return max(0, progress)

    def _update_A_and_b(self):
        r"""
        Builds/updates A, b based on the recent solution on algorithm object.
        """
        self._update_iterative_props()
        super()._update_A_and_b()
        self._apply_sources()

    def _get_residual(self, x=None):
        r"""
        Calculates solution residual based on the given ``x`` based on the
        following formula:

            ``R = A * x - b``

        """
        if x is None:
            x = self.x
        return self.A * x - self.b

    def set_BC(self, pores=None, bctype=[], bcvalues=[], mode="add"):
        msg = "Source term already present in given pores, can't assign BCs"
        # Ensure that given pores do not have source terms already set
        try:
            for item in self["pore.source"].keys():
                if np.any(self["pore.source." + item][pores]):
                    raise Exception(msg)
        except KeyError:
            pass
        # Assign BCs if above check passes
        super().set_BC(pores=pores, bctype=bctype, bcvalues=bcvalues, mode=mode)



r"""
Generators
==========

This module contains a selection of functions that generate various network
topologies

"""

from ._cubic import cubic
from ._delaunay import delaunay
from ._gabriel import gabriel
from ._voronoi import voronoi
from ._voronoi_delaunay_dual import voronoi_delaunay_dual
from ._template import cubic_template
from ._fcc import fcc
from ._bcc import bcc
from . import tools



r"""
Collection of functions for import/export of  data
==================================================

This module contains functionality for importing and exporting data
between OpenPNM and other formats.

"""

from ._utils import *
from ._dict import project_to_dict
from ._vtk import project_to_vtk
from ._pandas import project_to_pandas, network_to_pandas
from ._csv import project_to_csv, network_to_csv, network_from_csv
from ._hdf5 import project_to_hdf5, print_hdf5
from ._xdmf import project_to_xdmf
from ._marock import network_from_marock
from ._porespy import network_from_porespy
from ._paraview import project_to_paraview
from ._networkx import network_to_networkx, network_from_networkx
from ._statoil import network_from_statoil
from ._pergeos import network_to_pergeos, network_from_pergeos
from ._jsongraph import network_to_jsongraph, network_from_jsongraph
from ._stl import network_to_stl
from ._comsol import network_to_comsol
from ._salome import network_to_salome



import logging

from openpnm.models.collections.phase import standard_gas, standard_liquid
from openpnm.phase import Phase, _fetch_chemical_props

logger = logging.getLogger(__name__)


__all__ = [
    'Species',
    'StandardGas',
    'StandardLiquid',
]


class Species(Phase):
    r"""
    A special Phase object that represents a single species in a mixture

    This class provides a ``mixture`` attribute which allows one to lookup
    which mixture the species is associated with.

    Parameters
    ----------
    network : GenericNetwork
        The network to which this phase object will be attached.
    species : str, optional
        If provided, this is used to lookup tabulated constants from the
        ``chemicals`` package, which attempts to find a match. For instance,
        'water', 'Water', and 'H2O' all work. The contants are stored in
        ``species.params`` like ``species.params['molecular_weight']``. They
        can also be accessed using ``species['param.molecular_weight']`` using
        some behind the scenes python magic.
    name : str, optional
        The name of the phase. This is useful to keep track of the objects
        throughout the simulation.  The name must be unique to the project.
        If no name is given, one is generated.

    """

    def __init__(self, species, **kwargs):
        # Create temp first to ensure all look-ups pass before initializing obj
        from thermo import Chemical
        temp = _fetch_chemical_props(Chemical(species))
        super().__init__(**kwargs)
        self.params.update(temp)

    @property
    def mixture(self):
        for item in self.project:
            if hasattr(item, 'components'):
                for comp in item.components.values():
                    if self is comp:
                        return item
        logger.warn("No mixture phase found for this species")


class StandardLiquid(Species):
    """A ``Species`` object with built-in standard liquid-phase models."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_model_collection(standard_liquid)
        self.regenerate_models()


class StandardGas(Species):
    """A ``Species`` object with built-in standard gas-phase models."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_model_collection(standard_gas)
        self.regenerate_models()



import numpy as np
from openpnm.utils import PrintableList


__all__ = [
    'ParserMixin',
    'LabelMixin',
]


class ParserMixin:

    def _parse_indices(self, indices):
        r"""
        This private method accepts a list of pores or throats and returns a
        properly structured Numpy array of indices.

        Parameters
        ----------
        indices : int or array_like
            This argument can accept numerous different data types including
            boolean masks, integers and arrays.

        Returns
        -------
        A Numpy array of indices.

        Notes
        -----
        This method should only be called by the method that is actually using
        the locations, to avoid calling it multiple times.

        """
        if indices is None:
            indices = np.array([], ndmin=1, dtype=int)
        locs = np.array(indices, ndmin=1)
        # If boolean array, convert to indices
        if locs.dtype == bool:
            if np.size(locs) == self.Np:
                locs = self.Ps[locs]
            elif np.size(locs) == self.Nt:
                locs = self.Ts[locs]
            else:
                raise Exception('Mask of locations must be either '
                                + 'Np nor Nt long')
        locs = locs.astype(dtype=int)
        return locs

    def _parse_element(self, element, single=False):
        r"""
        This private method is used to parse the keyword \'element\' in many
        of the above methods.

        Parameters
        ----------
        element : str or List[str]
            The element argument to check.  If is None is recieved, then a list
            containing both \'pore\' and \'throat\' is returned.
        single : bool (default is False)
            When set to True only a single element is allowed and it will also
            return a string containing the element.

        Returns
        -------
        When ``single`` is ``False`` (default) a list containing the element(s)
        is returned.  When ``single`` is ``True`` a bare string containing the
        element is returned.

        """
        if element is None:
            element = ['pore', 'throat']
        # Convert element to a list for subsequent processing
        if isinstance(element, str):
            element = [element]
        # Convert 'pore.prop' and 'throat.prop' into just 'pore' and 'throat'
        element = [item.split('.', 1)[0] for item in element]
        # Make sure all are lowercase
        element = [item.lower() for item in element]
        # Deal with an plurals
        element = [item.rsplit('s', maxsplit=1)[0] for item in element]
        for item in element:
            if item not in ['pore', 'throat']:
                raise Exception('All keys must start with either pore or throat')
        # Remove duplicates if any
        _ = [element.remove(L) for L in element if element.count(L) > 1]
        if single:
            if len(element) > 1:
                raise Exception('Both elements recieved when single element '
                                + 'allowed')
            element = element[0]
        return element

    def _parse_labels(self, labels, element):
        r"""
        This private method is used for converting \'labels\' to a proper
        format, including dealing with wildcards (\*).

        Parameters
        ----------
        labels : str or List[str]
            The label or list of labels to be parsed. Note that the \* can be
            used as a wildcard.

        Returns
        -------
        A list of label strings, with all wildcard matches included if
        applicable.

        """
        if labels is None:
            raise Exception('Labels cannot be None')
        if isinstance(labels, str):
            labels = [labels]
        # Parse the labels list
        parsed_labels = []
        for label in labels:
            # Remove element from label, if present
            if element in label:
                label = label.split('.', 1)[-1]
            # Deal with wildcards
            if '*' in label:
                Ls = [L.split('.', 1)[-1] for L in self.labels(element=element)]
                if label.startswith('*'):
                    temp = [L for L in Ls if L.endswith(label.strip('*'))]
                if label.endswith('*'):
                    temp = [L for L in Ls if L.startswith(label.strip('*'))]
                temp = [element+'.'+L for L in temp]
            elif element+'.'+label in self.keys():
                temp = [element+'.'+label]
            else:
                temp = [element+'.'+label]
            parsed_labels.extend(temp)
            # Remove duplicates if any
            _ = [parsed_labels.remove(L) for L in parsed_labels
                 if parsed_labels.count(L) > 1]
        return parsed_labels

    def _parse_mode(self, mode, allowed=None, single=False):
        r"""
        This private method is for checking the \'mode\' used in the calling
        method.

        Parameters
        ----------
        mode : str or List[str]
            The mode(s) to be parsed
        allowed : List[str]
            A list containing the allowed modes.  This list is defined by the
            calling method.  If any of the received modes are not in the
            allowed list an exception is raised.
        single : bool (default is False)
            Indicates if only a single mode is allowed.  If this argument is
            True than a string is returned rather than a list of strings, which
            makes it easier to work with in the caller method.

        Returns
        -------
        A list containing the received modes as strings, checked to ensure they
        are all within the allowed set (if provoided).  Also, if the ``single``
        argument was True, then a string is returned.

        """
        if isinstance(mode, str):
            mode = [mode]
        for item in mode:
            if (allowed is not None) and (item not in allowed):
                raise Exception('\'mode\' must be one of the following: '
                                + allowed.__str__())
        # Remove duplicates, if any
        _ = [mode.remove(L) for L in mode if mode.count(L) > 1]
        if single:
            if len(mode) > 1:
                raise Exception('Multiple modes received when only one mode '
                                + 'is allowed by this method')
            mode = mode[0]
        return mode

    def _parse_prop(self, propname, element):
        element = self._parse_element(element, single=True)
        if propname.split('.', 1)[0] in ['pore', 'throat']:
            propname = propname.split('.', 1)[-1]
        return element + '.' + propname


class LabelMixin:
    """r
    This mixin adds functionality to the Base2 class so that boolean arrays
    are treated as labels
    """

    def _get_labels(self, element, locations, mode):
        r"""
        This is the actual label getter method, but it should not be called
        directly.  Use ``labels`` instead.
        """
        # Parse inputs
        locations = self._parse_indices(locations)
        element = self._parse_element(element=element)
        # Collect list of all pore OR throat labels
        labels = [i for i in self.keys(mode='labels') if i.split('.', 1)[0] in element]
        labels.sort()
        labels = np.array(labels)  # Convert to ndarray for following checks
        # Make an 2D array with locations in rows and labels in cols
        arr = np.vstack([self[item][locations] for item in labels]).T
        num_hits = np.sum(arr, axis=0)  # Number of locations with each label
        if mode in ['or', 'union', 'any']:
            temp = labels[num_hits > 0]
        elif mode in ['and', 'intersection']:
            temp = labels[num_hits == locations.size]
        elif mode in ['xor', 'exclusive_or']:
            temp = labels[num_hits == 1]
        elif mode in ['nor', 'not', 'none']:
            temp = labels[num_hits == 0]
        elif mode in ['nand']:
            temp = labels[num_hits == (locations.size - 1)]
        elif mode in ['xnor', 'nxor']:
            temp = labels[num_hits > 1]
        else:
            raise Exception('Unrecognized mode:'+str(mode))
        return PrintableList(temp)

    def labels(self, pores=[], throats=[], element=None, mode='union'):
        r"""
        Returns a list of labels present on the object

        Additionally, this function can return labels applied to a specified
        set of pores or throats

        Parameters
        ----------
        element : str
            Controls whether pore or throat labels are returned.  If empty then
            both are returned (default).
        pores (or throats) : array_like
            The pores (or throats) whose labels are sought.  If left empty a
            list containing all pore and throat labels is returned.
        mode : str, optional
            Controls how the query should be performed.  Only applicable
            when ``pores`` or ``throats`` are specified:

            ==============  ===================================================
            mode            meaning
            ==============  ===================================================
            'or'            Returns the labels that are assigned to *any* of
                            the given locations. Also accepts 'union' and 'any'
            'and'           Labels that are present on all the given locations.
                            also accepts 'intersection' and 'all'
            'xor'           Labels that are present on *only one*
                            of the given locations.Also accepts 'exclusive_or'
            'nor'           Labels that are *not* present on any of
                            the given locations. Also accepts 'not' and 'none'
            'nand'          Labels that are present on *all but one* of the
                            given locations
            'xnor'          Labels that are present on *more than one* of the
                            given locations.
            ==============  ===================================================

        Returns
        -------
        A list containing the labels on the object.  If ``pores`` or
        ``throats`` are given, the results are filtered according to the
        specified ``mode``.

        See Also
        --------
        props
        keys

        Notes
        -----
        Technically, *'nand'* and *'xnor'* should also return pores with *none*
        of the labels but these are not included.  This makes the returned list
        more useful.

        """
        # Short-circuit query when no pores or throats are given
        if (np.size(pores) == 0) and (np.size(throats) == 0):
            if element is None:
                element = ['pore', 'throat']
            if isinstance(element, str):
                element = [element]
            labels = PrintableList()
            for k, v in self.items():
                el, prop = k.split('.', 1)
                if (el in element) and (v.dtype == bool) and not prop.startswith('_'):
                    labels.append(k)
        elif (np.size(pores) > 0) and (np.size(throats) > 0):
            raise Exception('Cannot perform label query on pores and '
                            + 'throats simultaneously')
        elif np.size(pores) > 0:
            labels = self._get_labels(element='pore', locations=pores,
                                      mode=mode)
        elif np.size(throats) > 0:
            labels = self._get_labels(element='throat', locations=throats,
                                      mode=mode)
        return sorted(labels)

    def set_label(self, label, pores=None, throats=None, mode='add'):
        r"""
        Creates or updates a label array

        Parameters
        ----------
        label : str
            The label to apply to the specified locations
        pores : array_like
            A list of pore indices or a boolean mask of where given label
            should be added or removed (see ``mode``)
        throats : array_like
            A list of throat indices or a boolean mask of where given label
            should be added or removed (see ``mode``)
        mode : str
            Controls how the labels are handled.  Options are:

            =========== ======================================================
            mode        description
            =========== ======================================================
            'add'       (default) Adds the given label to the specified
                        locations while keeping existing labels

            'overwrite' Removes existing label from all locations before
                        adding the label in the specified locations

            'remove'    Removes the given label from the specified locations
                        leaving the remainder intact

            'purge'     Removes the specified label from the object completely.
                        This ignores the ``pores`` and ``throats`` arguments.

            'clear'     Sets all the labels to ``False`` but does not remove
                        the label array
            =========== ======================================================

        """
        self._parse_mode(mode=mode,
                         allowed=['add', 'overwrite', 'remove', 'purge',
                                  'clear'])

        if label.split('.', 1)[0] in ['pore', 'throat']:
            label = label.split('.', 1)[1]

        if (pores is not None) and (throats is not None):
            self.set_label(label=label, pores=pores, mode=mode)
            self.set_label(label=label, throats=throats, mode=mode)
            return
        elif pores is not None:
            locs = self._parse_indices(pores)
            element = 'pore'
        elif throats is not None:
            locs = self._parse_indices(throats)
            element = 'throat'

        if mode == 'add':
            if element + '.' + label not in self.keys():
                self[element + '.' + label] = False
            self[element + '.' + label][locs] = True
        if mode == 'overwrite':
            self[element + '.' + label] = False
            self[element + '.' + label][locs] = True
        if mode == 'remove':
            self[element + '.' + label][locs] = False
        if mode == 'clear':
            self['pore' + '.' + label] = False
            self['throat' + '.' + label] = False
        if mode == 'purge':
            _ = self.pop('pore.' + label, None)
            _ = self.pop('throat.' + label, None)

    def _get_indices(self, element, labels, mode='or'):
        r"""
        This is the actual method for getting indices, but should not be called
        directly.  Use ``pores`` or ``throats`` instead.
        """
        # Parse and validate all input values.
        element = self._parse_element(element, single=True)
        labels = self._parse_labels(labels=labels, element=element)

        # Begin computing label array
        if mode in ['or', 'any', 'union']:
            union = np.zeros([self._count(element), ], dtype=bool)
            for item in labels:  # Iterate over labels and collect all indices
                union = union + self[element+'.'+item.split('.', 1)[-1]]
            ind = union
        elif mode in ['and', 'all', 'intersection']:
            intersect = np.ones([self._count(element), ], dtype=bool)
            for item in labels:  # Iterate over labels and collect all indices
                intersect = intersect*self[element+'.'+item.split('.', 1)[-1]]
            ind = intersect
        elif mode in ['xor', 'exclusive_or']:
            xor = np.zeros([self._count(element), ], dtype=int)
            for item in labels:  # Iterate over labels and collect all indices
                info = self[element+'.'+item.split('.', 1)[-1]]
                xor = xor + np.int8(info)
            ind = (xor == 1)
        elif mode in ['nor', 'not', 'none']:
            nor = np.zeros([self._count(element), ], dtype=int)
            for item in labels:  # Iterate over labels and collect all indices
                info = self[element+'.'+item.split('.', 1)[-1]]
                nor = nor + np.int8(info)
            ind = (nor == 0)
        elif mode in ['nand']:
            nand = np.zeros([self._count(element), ], dtype=int)
            for item in labels:  # Iterate over labels and collect all indices
                info = self[element+'.'+item.split('.', 1)[-1]]
                nand = nand + np.int8(info)
            ind = (nand < len(labels)) * (nand > 0)
        elif mode in ['xnor', 'nxor']:
            xnor = np.zeros([self._count(element), ], dtype=int)
            for item in labels:  # Iterate over labels and collect all indices
                info = self[element+'.'+item.split('.', 1)[-1]]
                xnor = xnor + np.int8(info)
            ind = (xnor > 1)
        else:
            raise Exception('Unsupported mode: '+mode)
        # Extract indices from boolean mask
        ind = np.where(ind)[0]
        ind = ind.astype(dtype=int)
        return ind

    def pores(self, labels=None, mode='or', asmask=False):
        r"""
        Returns pore indicies where given labels exist, according to the logic
        specified by the ``mode`` argument.

        Parameters
        ----------
        labels : str or list[str]
            The label(s) whose pores locations are requested.  This argument
            also accepts '*' for wildcard searches.
        mode : str
            Specifies how the query should be performed.  The options are:

            ==============  ===================================================
            mode            meaning
            ==============  ===================================================
            'or'            Returns the labels that are assigned to *any* of
                            the given locations. Also accepts 'union' and 'any'
            'and'           Labels that are present on all the given locations.
                            also accepts 'intersection' and 'all'
            'xor'           Labels that are present on *only one*
                            of the given locations.Also accepts 'exclusive_or'
            'nor'           Labels that are *not* present on any of
                            the given locations. Also accepts 'not' and 'none'
            'nand'          Labels that are present on *all but one* of the
                            given locations
            'xnor'          Labels that are present on *more than one* of the
                            given locations.
            ==============  ===================================================

        asmask : bool
            If ``True`` then a boolean array of length Np is returned with
            ``True`` values indicating the pores that satisfy the query.

        Returns
        -------
        A Numpy array containing pore indices filtered by the logic specified
        in ``mode``.

        See Also
        --------
        throats

        Notes
        -----
        Technically, *nand* and *xnor* should also return pores with *none* of
        the labels but these are not included.  This makes the returned list
        more useful.

        To perform more complex or compound queries, you can opt to receive
        the result a a boolean mask (``asmask=True``), then manipulate the
        arrays manually.

        """
        if labels is None:
            labels = self.name
        ind = self._get_indices(element='pore', labels=labels, mode=mode)
        if asmask:
            ind = self.to_mask(pores=ind)
        return ind

    def throats(self, labels=None, mode='or', asmask=False):
        r"""
        Returns throat locations where given labels exist, according to the
        logic specified by the ``mode`` argument.

        Parameters
        ----------
        labels : str or list[str]
            The throat label(s) whose locations are requested.  If omitted,
            'all' throat inidices are returned.  This argument also accepts
            '*' for wildcard searches.
        mode : str
            Specifies how the query should be performed. The options are:

            ==============  ===================================================
            mode            meaning
            ==============  ===================================================
            'or'            Returns the labels that are assigned to *any* of
                            the given locations. Also accepts 'union' and 'any'
            'and'           Labels that are present on all the given locations.
                            also accepts 'intersection' and 'all'
            'xor'           Labels that are present on *only one*
                            of the given locations.Also accepts 'exclusive_or'
            'nor'           Labels that are *not* present on any of
                            the given locations. Also accepts 'not' and 'none'
            'nand'          Labels that are present on *all but one* of the
                            given locations
            'xnor'          Labels that are present on *more than one* of the
                            given locations.
            ==============  ===================================================

        asmask : bool
            If ``True`` then a boolean array of length Nt is returned with
            ``True`` values indicating the throats that satisfy the query.

        Returns
        -------
        A Numpy array containing throat indices filtered by the logic specified
        in ``mode``.

        See Also
        --------
        pores

        """
        if labels is None:
            labels = self.name
        ind = self._get_indices(element='throat', labels=labels, mode=mode)
        if asmask:
            ind = self.to_mask(throats=ind)
        return ind

    def filter_by_label(self, pores=[], throats=[], labels=None, mode='or'):
        r"""
        Returns which of the supplied pores (or throats) has the specified
        label(s)

        Parameters
        ----------
        pores, or throats : array_like
            List of pores or throats to be filtered
        labels : list of strings
            The labels to apply as a filter
        mode : str
            Controls how the filter is applied. The default value is
            'or'. Options include:

            ==============  ===================================================
            mode            meaning
            ==============  ===================================================
            'or'            Returns the labels that are assigned to *any* of
                            the given locations. Also accepts 'union' and 'any'
            'and'           Labels that are present on all the given locations.
                            also accepts 'intersection' and 'all'
            'xor'           Labels that are present on *only one*
                            of the given locations.Also accepts 'exclusive_or'
            'nor'           Labels that are *not* present on any of
                            the given locations. Also accepts 'not' and 'none'
            'nand'          Labels that are present on *all but one* of the
                            given locations
            'xnor'          Labels that are present on *more than one* of the
                            given locations.
            ==============  ===================================================

        Returns
        -------
        A list of pores (or throats) that have been filtered according the
        given criteria.  The returned list is a subset of the received list of
        pores (or throats).

        See Also
        --------
        pores
        throats

        """
        # Convert inputs to locations and element
        if (np.size(throats) > 0) and (np.size(pores) > 0):
            raise Exception('Can only filter either pores OR labels')
        if np.size(pores) > 0:
            element = 'pore'
            locations = self._parse_indices(pores)
        elif np.size(throats) > 0:
            element = 'throat'
            locations = self._parse_indices(throats)
        else:
            return np.array([], dtype=int)
        labels = self._parse_labels(labels=labels, element=element)
        labels = [element+'.'+item.split('.', 1)[-1] for item in labels]
        all_locs = self._get_indices(element=element, labels=labels, mode=mode)
        mask = self._tomask(indices=all_locs, element=element)
        ind = mask[locations]
        return locations[ind]

    def num_pores(self, labels='all', mode='or'):
        r"""
        Returns the number of pores of the specified labels

        Parameters
        ----------
        labels : list of strings, optional
            The pore labels that should be included in the count.
            If not supplied, all pores are counted.
        labels : list of strings
            Label of pores to be returned
        mode : str, optional
            Specifies how the count should be performed. The options are:

            ==============  ===================================================
            mode            meaning
            ==============  ===================================================
            'or'            Returns the labels that are assigned to *any* of
                            the given locations. Also accepts 'union' and 'any'
            'and'           Labels that are present on all the given locations.
                            also accepts 'intersection' and 'all'
            'xor'           Labels that are present on *only one*
                            of the given locations.Also accepts 'exclusive_or'
            'nor'           Labels that are *not* present on any of
                            the given locations. Also accepts 'not' and 'none'
            'nand'          Labels that are present on *all but one* of the
                            given locations
            'xnor'          Labels that are present on *more than one* of the
                            given locations.
            ==============  ===================================================

        Returns
        -------
        Np : int
            Number of pores with the specified labels

        See Also
        --------
        num_throats
        count

        Notes
        -----
        Technically, *'nand'* and *'xnor'* should also count pores with *none*
        of the labels, however, to make the count more useful these are not
        included.

        """
        # Count number of pores of specified type
        Ps = self._get_indices(labels=labels, mode=mode, element='pore')
        Np = np.shape(Ps)[0]
        return Np

    def num_throats(self, labels='all', mode='union'):
        r"""
        Return the number of throats of the specified labels

        Parameters
        ----------
        labels : list of strings, optional
            The throat labels that should be included in the count.
            If not supplied, all throats are counted.
        mode : str, optional
            Specifies how the count should be performed.  The options are:

            ==============  ===================================================
            mode            meaning
            ==============  ===================================================
            'or'            Returns the labels that are assigned to *any* of
                            the given locations. Also accepts 'union' and 'any'
            'and'           Labels that are present on all the given locations.
                            also accepts 'intersection' and 'all'
            'xor'           Labels that are present on *only one*
                            of the given locations.Also accepts 'exclusive_or'
            'nor'           Labels that are *not* present on any of
                            the given locations. Also accepts 'not' and 'none'
            'nand'          Labels that are present on *all but one* of the
                            given locations
            'xnor'          Labels that are present on *more than one* of the
                            given locations.
            ==============  ===================================================

        Returns
        -------
        Nt : int
            Number of throats with the specified labels

        See Also
        --------
        num_pores
        count

        Notes
        -----
        Technically, *'nand'* and *'xnor'* should also count throats with
        *none* of the labels, however, to make the count more useful these are
        not included.

        """
        # Count number of pores of specified type
        Ts = self._get_indices(labels=labels, mode=mode, element='throat')
        Nt = np.shape(Ts)[0]
        return Nt



"""
Tools
-----

"""
import numpy as np
from openpnm._skgraph import tools
import scipy.spatial as sptl


__all__ = [
    'parse_points',
    'add_all_label',
    'label_faces_cubic',
    'template_sphere_shell',
    'template_cylinder_annulus',
    'generate_base_points',
    'reflect_base_points',
    'get_centroid',
    'lloyd_relaxation',
]


def lloyd_relaxation(vor, mode='rigorous'):
    r"""
    Computes the center of mass for each polyhedra in a Voronoi tessellaion

    Parameters
    ----------
    vor : Scipy Voronoi object
        The object returned by ``scipy.spatial.Voronoi``
    mode : str
        Options are:

        =========== ================================================================
        mode        description
        =========== ================================================================
        rigorous    Computes a Delaunay triangulation of the input points then finds
                    the true centroid by computing the area/volume of each triangle/
                    tetrahedron to find the weighted average center of mass.
        fast        Computes the basic average of the input points without
                    accounting for the distribution of points. This is a decent
                    approximation and is *much* faster.
        =========== ================================================================

    """
    pts = vor.points
    for i, r in enumerate(vor.point_region):
        if np.all(np.array(vor.regions[r]) > 0):
            pts[i, :pts.shape[1]] = \
                get_centroid(vor.vertices[vor.regions[r]], mode=mode)
    if pts.shape[1] == 2:
        pts = np.c_[pts, np.zeros_like(pts[:, 0])]
    return pts


def get_centroid(pts, mode='rigorous'):
    r"""
    Finds the centroid of a given set of points

    Parameters
    ----------
    pts : ndarray
        The points in 2D or 3D
    mode : str
        Options are:

        =========== ================================================================
        mode        description
        =========== ================================================================
        rigorous    Computes a Delaunay triangulation of the input points then finds
                    the true centroid by computing the area/volume of each triangle/
                    tetrahedron to find the weighted average center of mass.
        fast        Computes the basic average of the input points without
                    accounting for the distribution of points. This is a decent
                    approximation and is *much* faster.
        =========== ================================================================

    Returns
    -------
    pt : ndarray
        A single coordinate representing the center of mass of the input points

    """
    if mode == 'rigorous':
        tri = sptl.Delaunay(pts)
        CoM = center_of_mass(simplices=tri.simplices.astype(np.int_),
                             points=tri.points.astype(float))
    elif mode == 'fast':
        CoM = pts.mean(axis=0)
    return CoM


# This function is causing all kinds of problems on the CI.  It is working on
# Ubuntu, Py3.8, but failing on all others.  It is working locally, windows with
# Py3.9.  jit and njit both fail for different reasons.  Anyway, the slow part
# is the tessellation of the points, not this actually CoM calculation
# so I'm commenting out the decorator for now.
# @njit
def center_of_mass(simplices, points):
    A = []
    centroids = np.zeros((simplices.shape[0], points.shape[1]), dtype=float)
    for i, s in enumerate(simplices):
        xy = points[s]
        cen = np.sum(points[s], axis=0)/simplices.shape[1]
        centroids[i, :] = cen.astype(float)
        if xy.shape[1] == 2:  # In 2D
            # Use Heron's formula to find area of arbitrary triangle
            a = np.sqrt((xy[0, 0] - xy[1, 0])**2 + (xy[0, 1] - xy[1, 1])**2)
            b = np.sqrt((xy[1, 0] - xy[2, 0])**2 + (xy[1, 1] - xy[2, 1])**2)
            c = np.sqrt((xy[2, 0] - xy[0, 0])**2 + (xy[2, 1] - xy[0, 1])**2)
            s = (a + b + c)/2
            temp = np.sqrt(s*(s-a)*(s-b)*(s-c))
        else:
            # Using formula from here: https://en.wikipedia.org/wiki/Tetrahedron
            d = np.concatenate((xy.T, np.ones((1, 4))))  # In 3D
            temp = np.abs(np.linalg.det(d))/6.0
            # temp = 0
        A.append(temp)
    A = np.array(A, dtype=float)
    CoM = np.array([(centroids[:, i]*A/A.sum()*len(A)).mean()
                    for i in range(centroids.shape[1])], dtype=float)
    return CoM


def parse_points(shape, points, reflect=False, f=1):
    r"""
    Converts given points argument to consistent format

    Parameters
    ----------
    shape : array_like
        The shape of the domain
    points : int or array_like
        If a scalar value then this indicates the number of points to generate. If
        an array, then these points are used directly.
    reflect : bool, optional
        If ``True`` the base points are reflected across the borders of the domain.
    f : float
        The fraction of points which should be reflected.  The default is 1 which
        reflects all the points in the domain, but this can lead to a lot of
        unnecessary points, so setting to 0.1 or 0.2 helps speed, but risks that
        the tessellation may not have smooth faces if not enough points are
        reflected.

    Returns
    -------
    points : ndarray
        The points after being cleaned and parsed correctly

    """
    # Deal with input arguments
    shape = np.array(shape, dtype=float)
    if isinstance(points, int):
        points = generate_base_points(num_points=points,
                                      domain_size=shape,
                                      reflect=False)
    else:
        points = np.array(points)
        # Add 3rd column to 2D points
        if points.shape[1] == 2:
            zeros = np.atleast_2d(np.zeros_like(points[:, 0])).T
            points = np.hstack((points, zeros))
        # Ensure z-axis is all 0's if shape ends in 0 (ie. square or disk)
        if shape[-1] == 0:
            points[:, -1] = 0.0
    if reflect:
        if np.any(tools.isoutside(points, shape=shape)):
            raise Exception('Some points lie outside the domain, '
                            + 'cannot safely apply reflection')
        if len(shape) == 3:
            points = reflect_base_points(points=points, domain_size=shape, f=f)
        elif len(shape) == 2:
            # Convert xyz to cylindrical, and back
            R, Q, Z = tools.cart2cyl(*points.T)
            R, Q, Z = reflect_base_points(np.vstack((R, Q, Z)), domain_size=shape, f=f)
            # Convert back to cartesean coordinates
            points = np.vstack(tools.cyl2cart(R, Q, Z)).T
        elif len(shape) == 1:
            # Convert to spherical coordinates
            R, Q, P = tools.cart2sph(*points.T)
            R, Q, P = reflect_base_points(np.vstack((R, Q, P)), domain_size=shape, f=f)
            # Convert to back to cartesean coordinates
            points = np.vstack(tools.sph2cart(R, Q, P)).T
    return points


def add_all_label(network):
    r"""
    Add 'node.all' and 'edge.all' to network dictionary

    Parameters
    ----------
    network : dict
        The network dictionary, containing 'node.coords' and 'edge.conns'

    Returns
    -------
    network : dict
        The supplied dictionary with the 'all' labels added

    Notes
    -----
    This function is helpful for working with OpenPNM

    """
    node_prefix = tools.get_node_prefix(network)
    edge_prefix = tools.get_edge_prefix(network)
    coords = network[node_prefix+'.coords']
    conns = network[edge_prefix+'.conns']
    network['pore.all'] = np.ones(coords.shape[0], dtype=bool)
    network['throat.all'] = np.ones(conns.shape[0], dtype=bool)
    return network


def label_faces_cubic(network, rtol=0.0):
    r"""
    Label the nodes sitting on the faces of the domain assuming the domain
    is cubic

    Parameters
    ----------
    network : dict
        The network dictionary contain 'node.coords'
    rtol : float
        Controls how closely a node must be to a face to be counted. It is
        computed relative to the fraction of domain size, as:
        ``hi_label = abs(1 - x[i]/x.max()) < rtol`` and
        ``lo_label = abs(x[i]/x.max()) < rtol``

    Returns
    -------
    network : dict
        The network dictionary with the face labels added

    """
    node_prefix = tools.get_node_prefix(network)
    coords = network[node_prefix+'.coords']
    dims = tools.dimensionality(network)
    coords = np.around(coords, decimals=10)
    min_labels = ['left', 'front', 'bottom']
    max_labels = ['right', 'back', 'top']
    min_coords = np.amin(coords, axis=0)
    max_coords = np.amax(coords, axis=0)
    for ax in np.where(dims)[0]:
        network[node_prefix + '.' + min_labels[ax]] = \
            abs((coords[:, ax]-min_coords[ax])/max_coords[ax]) <= rtol
        network[node_prefix + '.' + max_labels[ax]] = \
            abs(1-coords[:, ax]/max_coords[ax]) <= rtol
    return network


def _template_sphere_disc(dim, outer_radius, inner_radius):
    r"""
    This private method generates an image array of a sphere/shell-disc/ring.

    It is useful for passing to Cubic networks as a ``template`` to make
    networks with desired shapes.

    Parameters
    ----------
    dim : int
        Network dimension
    outer_radius : int
        Number of the nodes in the outer radius of the network
    inner_radius : int
        Number of the nodes in the inner radius of the network

    Returns
    -------
    im : array_like
        A Numpy array containing 1's to demarcate the desired shape, and 0's
        elsewhere.

    """
    rmax = np.array(outer_radius, ndmin=1)
    rmin = np.array(inner_radius, ndmin=1)
    ind = 2 * rmax - 1
    coord = np.indices((ind * np.ones(dim, dtype=int)))
    coord = coord - (ind - 1)/2
    x = coord[0, :]
    y = coord[1, :]
    if dim == 2:
        img = (x ** 2 + y ** 2) < rmax ** 2
    elif dim == 3:
        z = coord[2, :]
        img = (x ** 2 + y ** 2 + z ** 2) < rmax ** 2
    if rmin[0] != 0:
        if dim == 2:
            img_min = (x ** 2 + y ** 2) > rmin ** 2
        elif dim == 3:
            img_min = (x ** 2 + y ** 2 + z ** 2) > rmin ** 2
        img = img * img_min
    return img


def template_sphere_shell(r_outer, r_inner=0):
    r"""
    This method generates an image array of a sphere-shell.

    It is useful for passing to Cubic networks as a ``template`` to make
    spherical shaped networks.

    Parameters
    ----------
    r_outer : int
        Number of nodes in the outer radius of the sphere
    r_inner : int, optional
        Number of nodes in the inner radius of the shell.  A value of 0 will
        result in a solid sphere.

    Returns
    -------
    im : array_like
        A Numpy array containing 1's to demarcate the sphere-shell, and 0's
        elsewhere.

    """
    img = _template_sphere_disc(dim=3, outer_radius=r_outer,
                                inner_radius=r_inner)
    return img


def template_cylinder_annulus(z, r_outer, r_inner=0):
    r"""
    This method generates an image array of a disc-ring.

    It is useful for passing to Cubic networks as a ``template`` to make
    circular-shaped 2D networks.

    Parameters
    ----------
    z : int
        The height of the cylinder. A value of 0 will result in a circle
    r_outer : int
        Number of nodes in the outer radius of the cylinder
    r_inner : int, optional
        Number of the nodes in the inner radius of the annulus.  A value of 0
        will result in a solid cylinder.

    Returns
    -------
    im : array_like
        A Numpy array containing 1's to demarcate the disc-ring, and 0's
        elsewhere.

    """

    img = _template_sphere_disc(dim=2, outer_radius=r_outer,
                                inner_radius=r_inner)
    if z > 0:
        img = np.tile(np.atleast_3d(img), reps=z)
    return img


def reflect_base_points(points, domain_size, f=1):
    r"""
    Relects a set of points about the faces of a given domain

    Parameters
    ----------
    points : ndarray
        The coordinates of the points to be reflected.  The points should be
        in the coordinate system corresponding to the the domain.
    domain_size : list or array
        Controls the size and shape of the domain, as follows:

        ========== ============================================================
        shape      result
        ========== ============================================================
        [x, y, z]  Points will be reflected about all 6 faces
        [x, y, 0]  Points will be relfected about the x and y faces
        [r, z]     Points will be reflected above and below the end faces and
                   across the perimeter
        [r, 0]     Points will be reflected across the perimeter
        [r]        Points will be reflected across the outer surface
        ========== ============================================================

    f : float
        The fraction of points which should be reflected.  The default is 1 which
        reflects all the points in the domain, but this can lead to a lot of
        unnecessary points, so setting to 0.1 or 0.2 helps speed, but risks that
        the tessellation may not have smooth faces if not enough points are
        reflected.

    Returns
    -------
    points : ndarray
        The coordinates of the original points plus the new reflected points
    """
    assert 0 <= f <= 1, 'f must be between 0 and 1'
    domain_size = np.array(domain_size)
    if len(domain_size) == 1:
        r, theta, phi = points
        new_r = 2*domain_size[0] - r
        r = np.hstack([r, new_r])
        theta = np.hstack([theta, theta])
        phi = np.hstack([phi, phi])
        points = np.vstack((r, theta, phi))
        # Trim excess points outside radius
        hi = domain_size[0]*(1+f)
        keep = (points[0, :] <= hi)
        points = points[:, keep]
    if len(domain_size) == 2:
        r, theta, z = points
        new_r = 2*domain_size[0] - r
        r = np.hstack([r, new_r])
        theta = np.hstack([theta, theta])
        z = np.hstack([z, z])
        if domain_size[1] != 0:  # If not a disk
            r = np.hstack([r, r, r])
            theta = np.hstack([theta, theta, theta])
            z = np.hstack([z, -z, 2*domain_size[1]-z])
        points = np.vstack((r, theta, z))
        # Trim excess basepoints above and below cylinder
        hi = domain_size[1]*(1+f)
        lo = domain_size[1]*(-f)
        keep = (points[2, :] <= hi)*(points[2, :] >= lo)
        # Trim excess points outside radius
        hi = domain_size[0]*(1+f)
        keep *= (points[0, :] <= hi)
        points = points[:, keep]
    elif len(domain_size) == 3:
        Nx, Ny, Nz = domain_size
        # Reflect base points about all 6 faces
        orig_pts = points
        points = np.vstack((points,
                            [-1, 1, 1] * orig_pts + [2.0 * Nx, 0, 0]))
        points = np.vstack((points, [-1, 1, 1] * orig_pts))
        points = np.vstack((points,
                            [1, -1, 1] * orig_pts + [0, 2.0 * Ny, 0]))
        points = np.vstack((points, [1, -1, 1] * orig_pts))
        if domain_size[2] != 0:
            points = np.vstack((points,
                               [1, 1, -1] * orig_pts + [0, 0, 2.0 * Nz]))
            points = np.vstack((points, [1, 1, -1] * orig_pts))
        # Trim excess basepoints
        hi = domain_size*(1+f)
        lo = domain_size*(-f)
        keep = np.all(points <= hi, axis=1)
        keep *= np.all(points >= lo, axis=1)
        points = points[keep, :]
    return points


def generate_base_points(num_points, domain_size, reflect=True, f=1):
    r"""
    Generates a set of randomly distributed points in rectilinear coordinates
    for use in spatial tessellations

    The points can be distributed in spherical, cylindrical, or rectilinear
    domains, as well as 2D and 3D (disks and squares)

    Parameters
    ----------
    num_points : scalar
        The number of base points that lie within the domain
    domain_size : list or array
        Controls the size and shape of the domain, as follows:

        ========== ============================================================
        shape      result
        ========== ============================================================
        [x, y, z]  A 3D cubic domain of dimension x, y and z, with points in
                   the range [0:x, 0:y, 0:z]
        [x, y, 0]  A 2D square domain of size x by y, with points in the range
                   [0:x, 0:y, 0:0]
        [r, z]     A 3D cylindrical domain of radius r and height z, with
                   points in the range [-r:r, -r:r, 0:z]
        [r, 0]     A 2D circular domain of radius r, with points in the range
                   [-r:r, -r:r, 0:0]
        [r]        A 3D spherical domain of radius r, with points in the range
                   [-r:r, -r:r, -r:r]
        ========== ============================================================

    reflect : bool
        If ``True``, the base points are generated as specified then reflected
        about each face of the domain.  This essentially tricks the
        tessellation functions into creating smooth faces at the boundaries
        once these excess points are trimmed. Note that the surface is not
        perfectly smooth for the curved faces.
    f : float
        The fraction of points which should be reflected.  The default is 1 which
        reflects all the points in the domain, but this can lead to a lot of
        unnecessary points, so setting to 0.1 or 0.2 helps speed, but risks that
        the tessellation may not have smooth faces if not enough points are
        reflected.

    Notes
    -----
    To convert between coordinate systems see ``cart2sph`` and ``cart2cyl``

    """
    if len(domain_size) == 1:  # Spherical
        shape = np.array([2*domain_size[0], 2*domain_size[0], 2*domain_size[0]])
        base_pts = np.random.rand(num_points*9, 3)  # Generate more than needed
        # Convert to spherical coordinates
        X, Y, Z = ((np.array(base_pts - [0.5, 0.5, 0.5]))*shape).T
        R, Q, P = tools.cart2sph(X, Y, Z)
        # Keep points outside the domain
        inds = R <= domain_size[0]
        R, Q, P = R[inds], Q[inds], P[inds]
        # Trim internal points to give requested final number
        R, Q, P = R[:num_points], Q[:num_points], P[:num_points]
        # Reflect base points across perimeter
        if reflect:
            R, Q, P = reflect_base_points(np.vstack((R, Q, P)), domain_size, f=f)
        # Convert to Cartesean coordinates
        base_pts = np.vstack(tools.sph2cart(R, Q, P)).T

    elif len(domain_size) == 2:  # Cylindrical or Disk
        shape = np.array([2*domain_size[0], 2*domain_size[0], domain_size[1]])
        base_pts = np.random.rand(num_points*9, 3)  # Generate more than needed
        # Convert to cylindrical coordinates
        X, Y, Z = ((np.array(base_pts - [0.5, 0.5, 0]))*shape).T
        R, Q, Z = tools.cart2cyl(X, Y, Z)
        # Trim points outside the domain
        inds = R <= domain_size[0]
        R, Q, Z = R[inds], Q[inds], Z[inds]
        # Reduce to requested number of points
        R, Q, Z = R[:num_points], Q[:num_points], Z[:num_points]
        if reflect:
            R, Q, Z = reflect_base_points(np.vstack((R, Q, Z)), domain_size, f=f)
        # Convert to Cartesean coordinates
        base_pts = np.vstack(tools.cyl2cart(R, Q, Z)).T

    elif len(domain_size) == 3:  # Cube or square
        base_pts = np.random.rand(num_points, 3)*domain_size
        if reflect:
            base_pts = reflect_base_points(base_pts, domain_size, f=f)

    return base_pts
